;This program is free software; you can redistribute it and/or modify it under
;the terms of the GNU General Public License version 2 as published by the
;Free Software Foundation.
;This program is distributed in the hope that it will be useful, but WITHOUT
;ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
;details.
;You should have received a copy of the GNU General Public License along with
;this program (in the file LICENSE.TXT); if not, write to the Free Software
;Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;
;REVISIONS:
;2016-09-27 e.s. Uppercase RTEST
;2016-01-08 e.s. Replace random number generator with XPL0-PC version.
; Add Seed address to SYSVEC table. Change autorun switch to '-R' for MP/M
; compatibility.
;2015-12-07 e.s. Multi-dim array support. I2L code no longer V4 compatible.
; Better random number generator.
;2015-11-29 e.s. Change opcode $1D to xor. Update INTIN HEXIN. Add opcodes
; FORDDO DEC shifts MKARY. Add intrinsics SYSVEC RESTORE TEXT0 GETREG CHKEY
; FNAME DELAY CLEAR CLREOL CURSOR NORMAL BOLD INSLINE DELLINE CRTINIT CRTEXIT.
; Devices renumbered. Add device 8. Move stack to high memory to fix overflow.
; Autorun (/R) switch. Replace MAKEFCB with SETFCB. Simplified EXTEND.
;2015-10-02 e.s. Fix program end address display. Add opcode $29.
;2015-10-01 e.s. Fix FORDO < > <= >=. Fix JPCDO to test all 16 bits.
; Change intrinsic 6 to RESTRT. Add opcodes $25..$28. Device 7
; uses filename on command-line. Show start/end hex address for
; CP/M 3 SAVE.
;

;	ASSEMBLE WITH CP/M ASM/MAC OR MICROSOFT M80

	ASEG	; FOR M80

;TITLE 'I2L INTERPRETER FOR THE 8080    12-15-80'
;.
;.		*** I2L Interpreter for the 8080 MPU ***
;.
;.			Copyright 1978 by Wayne Wall
;.
;	Modified and completed by Dana Trout, 1979
;  (Lines which have a comment starting with ;. are unchanged.)
;
; Code added by Dana Trout is protected by the following copyright
;			Copyright 1979, 1980 by Dana Trout
;
;  Version notes:
;	For code produced by V4D of the compiler
;	Has minimal intrinsics, devices and size
;
;  Error number list:
;	1  division by 0
;	2  no more space
;	3  IO error
;	4  invalid opcode
;	5  invalid function number
;	10 loader failure
;	11 call from wrong level
;	12 invalid device number
;
;  Devices in this version:
;	device 0 - console
;	device 1 - unbuffered console
;	device 2 - printer/reader
;	device 3 - floppy disk (no prompts)
;	device 4 - not implemented
;	device 5 - not implemented
;	device 6 - floppy disk (prompts for file name)
;	device 7 - null device
;	device 8 - pseudo device
;
;.  This is the I2L (Interpreted Intermediate Language) Interpreter
;.  for the 8080 (and Z80) microprocessors.  The I2L interpreter
;.  is the machine dependent runtime code which implements the
;.  psuedo-machine on a given hardware processor
;.
;.  Many thanks to Peter Boyle, the creator of 'XPL0', who is
;.  indirectly responsible for my writing this program.
;.


;	System dependant equates

MONITR	EQU	0000H	;system monitor
CPMIO	EQU	0005H	;system I/O entry point
;
;	Equates unique to CP/M
;
CPMIO	EQU	5	;CPM I/O entry point
DFCB	EQU	5CH	;default FCB
DDMA	EQU	80H	;default DMA addr.

GETCHR	EQU	1	;fcn. # to get one character from console
PUTCHR	EQU	2	;fcn. # to write one char to console
LSTPRN	EQU	5	;fcn. # to write to list device
GETSTAT	EQU	7	;fcn. # to get I/O status byte
PUTSTAT EQU	8	;fcn. # to set I/O status byte
BUFPRN	EQU	9	;fcn. # for buffered print
BUFINP	EQU	10	;fcn. # for buffered input
CONSTAT	EQU	11	;fcn. # to get console status
HEADUP	EQU	12	;fcn. # to lift head on cur drive
RESETD	EQU	13	;fcn. # to initialize BDOS, select
			;  drive A, and set DMA buf=80H
LOGOND	EQU	14	;fcn. # to log on drive
OPNFIL	EQU	15	;fcn. # for open file
CLSFIL	EQU	16	;fcn. # for close file
SERFIL	EQU	17	;fcn. # to search for file
NSERFIL	EQU	18	;fcn. # to search for next occurance
DELFIL	EQU	19	;fcn. # to delete file
READSEC	EQU	20	;fcn. # for read sector
WRITSEC	EQU	21	;fcn. # for write sector
MAKFIL	EQU	22	;fcn. # for create file
RENFIL	EQU	23	;fcn. # to rename FCB
GETLOG	EQU	24	;fcn. # to get login vector
GETDRV	EQU	25	;fcn. # to get current drive #
SETDMA	EQU	26	;fcn. # to set DMA address
GETALOC	EQU	27	;fcn. # to get alloc. vect.of cur drive
;
;	END of CP/M equates
;

;	Error code definitions

DIVERR	EQU	1	;division by zero
RMERR	EQU	2	;no more room
IOERR	EQU	3	;IO error
OPCERR	EQU	4	;invalid opcode encountered
ERRFUN	EQU	5	;function code is in error
LODERR	EQU	10	;loader error encountered
CALERR	EQU	11	;call from wrong level
DEVERR	EQU	12	;invalid device number given

HISPAC	EQU	0A000H	;highest available memory + 1

EOFCHR	EQU	1AH	;end of file (control z)
FFCHR	EQU	0CH	;form feed
CR	EQU	0DH	;carriage return
LF	EQU	0AH	;line feed
RUB	EQU	7FH	;rubout
BACKSP	EQU	8	;backspace
CTRLU	EQU	15H	;line abort

	ORG	100H

	JMP	ENTER

; Begin terminal installer area.
; Note: Not all fields are used by the I2L interpreter.

	ORG	0103H

	DB	'Z3ENV'		; Z-System signature		0103
	DB	1		; external environment		0108
Z3EADR:	DW	0		; ZENV address			0109

; The following 5 bytes are reserved exclusively for user applications

	DB	0,0,0,0,0	; user patch bytes		010B

; If the following byte is non-zero, a warm boot is NOT performed on
; exit to CP/M.

NOBOOT:	DB	0		; no warm boot flag		0110

; Identification and version

	DW	4683H		; id  ('DXF' in radix 36)	0111
	DB	0;REL		; release			0113
	DB	0;REV		; revision			0114

	ORG	0124H

; cpu speed (TURBO PASCAL compatible)

SPEED:	DW	4		; 1-8191 MHz 			0124

	ORG	014FH

; Arrow key codes (default = Wordstar style)

	DB	'E'-40H		; up arrow			014F
	DB	'X'-40H		; down arrow			0150
	DB	'D'-40H		; right arrow			0151
	DB	'S'-40H		; left arrow			0152

; Terminal install area (TURBO PASCAL compatible)

TERMN:	DB	20			; terminal name		0153
	DB	'Televideo 912/920/92'
TCOLS:	DB	80			; max cols		0168
TROWS:	DB	24			; max rows		0169
	DB	0			; reserved		016A
TINIT:	DB	0,0,0,0,0,0,0,0		; terminal init	seq.	016B
	DB	0,0,0,0,0,0,0,0
TEXIT:	DB	0,0,0,0,0,0,0,0		; terminal exit seq.	017B
	DB	0,0,0,0,0,0,0,0
TGXY:	DB	4,27,'=',0,0,0,0,0	; cursor position	018B
	DB	0,0,0,0,0,0,0,0
TISBIN:	DB	1			; decimal=0 binary=1	019B
TXOFS:	DB	32			; offset added to col	019C
TYOFS:	DB	32			; offset added to row	019D
TXPOS:	DB	4			; pos insert column	019E
TYPOS:	DB	3			; pos insert row	019F
TGXYD:	DW	0			; cursor pos delay 	01A0
TCLS:	DB	1,26,0,0,0,0		; clear screen		01A2
THOME:	DB	0,0,0,0,0,0		; home cursor		01A8
TINSLN:	DB	2,27,'E',0,0,0		; insert line		01AE
TDELLN:	DB	2,27,'R',0,0,0		; delete line		01B4
TCLSD:	DW	0			; clear screen delay 	01BA
TCEOL:	DB	2,27,'T',0,0,0		; clear line		01BC
THIVID:	DB	2,27,'(',0,0,0		; highlight/reverse	01C2
TLOVID:	DB	2,27,')',0,0,0		; normal video		01C8
TCEOLD:	DW	0			; clear line delay 	01CE

; End terminal installer area

;
;  First a few handy vectors - so that overlays can find things
;  easily.
;
ENTER:	JMP	LODST	;this will be changed by the loader
	JMP	LODST	;this won't be changed
	JMP	CMLRET	;return from intrinsics
	JMP	GETC	;get a byte
	JMP	OUTTO	;put a byte
	JMP	INIDEV	;open for input
	JMP	INODEV	;open for output
	JMP	CLODEV	;close a device
;
;	*** IO Device linkage table ***
;
	ORG	ENTER+20H	;give some room for more jumps
;
;  Device drivers linked into this table are direct subroutines.
;  They do the required op and then return with the carry flag
;  cleared if they succeeded or set if they failed.  All registers
;  may be used.  On output the accumulator contains the byte
;  to be output and on input it contains the returned byte.
;  The eighth bit of ASCII characters must be zero.  Device 0
;  is assumed to be the console which can never give an error.
;

IOTAB	EQU	$
;
;  Input addresses
;
	DW	KEYIN	;device 0 - console
	DW	KBORD	;device 1 - unbuffered console
	DW	NULDEV	;device 2 - printer/reader
	DW	BYFMCP	;device 3 - floppy disk (no prompts)
	DW	NULDEV	;device 4 -
	DW	NULDEV	;device 5 -
	DW	BYFMCP	;device 6 - floppy disk (prompts for file name)
	DW	NULDEV	;device 7 - null device
	DW	PSIN	;device 8 - pseudo device
;
;  Output addresses
;
	DW	TVOUT	;device 0 - console
	DW	CRT	;device 1 - unbuffered console
	DW	PRTOUT	;device 2 - printer/reader
	DW	BYTOCP	;device 3 - floppy disk (no prompts)
	DW	NULDEV	;device 4 -
	DW	NULDEV	;device 5 -
	DW	BYTOCP	;device 6 - floppy disk (prompts for file name)
	DW	NULDEV	;device 7 - null device
	DW	PSOUT	;device 8 - pseudo device
;
;  Close output file
;
	DW	NULDEV	;device 0 - console
	DW	NULDEV	;device 1 - unbuffered console
	DW	NULDEV	;device 2 - printer/reader
	DW	CLSCPM	;device 3 - floppy disk (no prompts)
	DW	NULDEV	;device 4 -
	DW	NULDEV	;device 5 -
	DW	CLSCPM	;device 6 - floppy disk (prompts for file name)
	DW	NULDEV	;device 7 - null device
	DW	NULDEV	;device 8 - pseudo device
;
;  Open input file
;
	DW	NULDEV	;device 0 - console
	DW	NULDEV	;device 1 - unbuffered console
	DW	NULDEV	;device 2 - printer/reader
	DW	FNDIT	;device 3 - floppy disk (no prompts)
	DW	NULDEV	;device 4 -
	DW	NULDEV	;device 5 -
	DW	OPNIN	;device 6 - floppy disk (prompts for file name)
	DW	NULDEV	;device 7 - null device
	DW	PSOPNI	;device 8 - pseudo device
;
;  Open output file
;
	DW	NULDEV	;device 0 - console
	DW	NULDEV	;device 1 - unbuffered console
	DW	NULDEV	;device 2 - printer/reader
	DW	OPNOUT1	;device 3 - floppy disk (no prompts)
	DW	NULDEV	;device 4 -
	DW	NULDEV	;device 5 -
	DW	OPNOUT	;device 6 - floppy disk (prompts for file name)
	DW	NULDEV	;device 7 - null device
	DW	PSOPNO	;device 8 - pseudo device

DEVMAX	EQU	($-IOTAB)/(5*2)	;maximum number of devices

;
;	Register image area for machine language calls
;
REGIM:	DS	1	; 0  FLAGS
	DS	1	; 1  A
	DS	1	; 2  B
	DS	1	; 3  C
	DS	1	; 4  D
	DS	1	; 5  E
	DS	1	; 6  H
	DS	1	; 7  L

;	POS counter for each device
;	This counter counts the number of characters output to
;	the device since the last carriage return.

POS0:	DW	0	;console
POS1:	DW	0
POS2:	DW	0
POS3:	DW	0	;floppy disk
POS4:	DW	0	;list printer
POS5:	DW	0
POS6:	DW	0	;unbuffered console
POS7:	DW	0	;floppy (no prompts)
POS8:	DW	0	;pseudo device

;	IO error bytes
;	There is one IO error byte for each function of each
;	device.  They are used to return specific error
;	information

; Input Error bytes
IER0:	DB	0
IER1:	DB	0
IER2:	DB	0
IER3:	DB	0
IER4:	DB	0
IER5:	DB	0
IER6:	DB	0
IER7:	DB	0
IER8:	DB	0

; Output Error bytes
OER0:	DB	0
OER1:	DB	0
OER2:	DB	0
OER3:	DB	0
OER4:	DB	0
OER5:	DB	0
OER6:	DB	0
OER7:	DB	0
OER8:	DB	0

; Close Output Error bytes
CER0:	DB	0
CER1:	DB	0
CER2:	DB	0
CER3:	DB	0
CER4:	DB	0
CER5:	DB	0
CER6:	DB	0
CER7:	DB	0
CER8:	DB	0

; Open Input Error bytes
OIER0:	DB	0
OIER1:	DB	0
OIER2:	DB	0
OIER3:	DB	0
OIER4:	DB	0
OIER5:	DB	0
OIER6:	DB	0
OIER7:	DB	0
OIER8:	DB	0

; Open Output Error bytes
OOER0:	DB	0
OOER1:	DB	0
OOER2:	DB	0
OOER3:	DB	0
OOER4:	DB	0
OOER5:	DB	0
OOER6:	DB	0
OOER7:	DB	0
OOER8:	DB	0

; System vector table
SVEC:	DW	RFCB	; 0  Read FCB address
	DW	WFCB	; 1  Write FCB address
	DW	REGIM	; 2  Register image address
	DW	ENTER	; 3  Jump table address
	DW	SEED	; 4  Current 32-bit RNG seed address
	DS	2	; 5  spare
	DS	2	; 6  spare
	DS	2	; 7  spare
	DS	2	; 8  spare
	DS	2	; 9  spare
	DS	2	; 10 spare
	DS	2	; 11 spare
	DS	2	; 12 spare
	DS	2	; 13 spare
	DS	2	; 14 spare
	DS	2	; 15 spare

;	Console input buffer

CIBUF:	DB	BUFE-INBUF-2 ;max number of characters allowed
NUMCAR:	DB	0	;number of characters gathered
INBUF:	DS	80	;input line buffer
BUFE	EQU	$

;.
;.	Dispatch Tables
;.
;.  The I2L opcode dispatch table.  Holds the address of the
;.  routine which processes the I2L instruction.  Indexed by
;.  I2L opcode * 2.

OPTAB:	DW	EXITDO	;.OP=0	exit interpreter
	DW	LODDO	;.OP=1	load an integer variable's value
	DW	LDXDO	;.OP=2	indexed load of a byte
	DW	STODO	;.OP=3	store top of stack into integer variable
	DW	STXDO	;.OP=4	indexed store of a byte
	DW	CALDO	;.OP=5	call an I2L procedure
	DW	RETDO	;.OP=6	return from an I2L procedure
	DW	JMPDO	;.OP=7	jump to an I2L procedure
	DW	JPCDO	;.OP=8	jump on FALSE
	DW	HPIDO	;.OP=9	increment heap pointer by arg
	DW	ARGDO	;.OP=A	get procedure arguments
	DW	IMMDO	;.OP=B	immediate load of integer constant
	DW	CMLDO	;.OP=C	call a machine language function
	DW	ADDDO	;.OP=D	add
	DW	SUBDO	;.OP=E	subtract
	DW	MUYDO	;.OP=F	multiply
	DW	DIVDO	;.OP=10	divide
	DW	NEGDO	;.OP=11	monadic minus
	DW	EQDO	;.OP=12	test for equal
	DW	NEDO	;.OP=13	test for not equal
	DW	GEDO	;.OP=14	test for >=
	DW	GTDO	;.OP=15	test for >
	DW	LEDO	;.OP=16	test for <=
	DW	LTDO	;.OP=17	test for <
	DW	FORDO	;.OP=18	FOR TO loop control
	DW	INCDO	;.OP=19	increment and push
	DW	ORDO	;.OP=1A	boolean OR
	DW	ANDDO	;.OP=1B	boolean AND
	DW	NOTDO	;.OP=1C	boolean COMPLEMENT
;**	DW	DUPCAT	;OP=1D	double TOS   *** REMOVED ***
	DW	XORDO	;OP=1D	boolean XOR
	DW	DBADO	;OP=1E	NOS+TOS*2
	DW	STDDO	;OP=1F	indirect save
	DW	DBXDO	;OP=20	indirect get
	DW	ADRDO	;OP=21	address of variable
	DW	LDIDO	;OP=22	indirect get
	DW	LDADO	;OP=23	absolute get
	DW	SIMMDO	;OP=24	short immediate
	DW	CAJMP	;OP=25	case jump
	DW	JSRDO	;OP=26	optimized call
	DW	RTSDO	;OP=27	optimized return
	DW	DRPDO	;OP=28	discard TOS
	DW	EXTDO	;OP=29	call external subroutine
;
	DW	FORDDO	;OP=2A	FOR DOWNTO loop control
	DW	DECDO	;OP=2B	decrement and push
	DW	OPERR	;OP=2C
	DW	OPERR	;OP=2D
	DW	OPERR	;OP=2E
	DW	OPERR	;OP=2F
	DW	OPERR	;OP=30
	DW	OPERR	;OP=31
	DW	OPERR	;OP=32
	DW	OPERR	;OP=33
	DW	OPERR	;OP=34
	DW	OPERR	;OP=35
	DW	OPERR	;OP=36
	DW	OPERR	;OP=37
	DW	OPERR	;OP=38
	DW	OPERR	;OP=39
	DW	OPERR	;OP=3A
	DW	OPERR	;OP=3B
	DW	MKARY	;OP=3C	reserve multidim array
	DW	ASRDO	;OP=3D	arithmetic right shift
	DW	LSLDO	;OP=3E	logical left shift
	DW	LSRDO	;OP=3F	logical right shift

;.
;.	Machine language function dispatch table
;.
;.  This table is used by the CML (Call Machine Language)
;.  instruction in order to dispatch to the proper machine
;.  code routine.  It is indexed by machine code function
;.  number.
;.

FUNTAB:	DW	ABSFUN	;FUN=40	#0  absolute value
	DW	RANFUN	;FUN=41	#1  random number
	DW	REMFUN	;FUN=42	#2  remainder (modulo)
	DW	RESERV	;FUN=43 #3  reserve array
	DW	SWAPFU	;FUN=44 #4  swap bytes
	DW	EXTEN	;FUN=45 #5  extend sign from low byte
	DW	RESTRT	;FUN=46 #6  restart program
;**	DW	SHORT	;FUN=46 #6  make one byte signed   *** REMOVED ***
	DW	CHIN	;FUN=47 #7  byte in
	DW	CHOUT	;FUN=48 #8  byte out
	DW	SKIP	;FUN=49 #9  write EOL
	DW	INTIN	;FUN=4A #10 read integer
	DW	INTOUT	;FUN=4B #11 write integer
	DW	TEXT	;FUN=4C #12 write string
	DW	IDEVIN	;FUN=4D #13 initialize input device
	DW	ODEVIN	;FUN=4E #14 initialize output device
	DW	ODEVCL	;FUN=4F #15 close an output device
	DW	ABORT	;FUN=50 #16 unconditional escape
;  The non-resident compiler needs ops up to here
	DW	CALLER	;FUN=51 #17 call external procedure
	DW	SPACE	;FUN=52 #18 determine remaining space
	DW	RERUN	;FUN=53 #19 test rerun flag
	DW	TSTSP	;FUN=54 #20 get current heap pointer
	DW	SETSP	;FUN=55 #21 set heap pointer
	DW	SETRUN	;FUN=56 #22 set the rerun flag
;  The resident compiler needs ops up to here
	DW	CURSOR	;FUN=57 #23 set cursor position
	DW	BADFUN	;FUN=58 #24 reserved
	DW	BADFUN	;FUN=59 #25
	DW	HEXIN	;FUN=5A #26 read hex integer
	DW	HEXPRN	;FUN=5B #27 write hex integer
	DW	BADFUN	;FUN=5C #28
	DW	BADFUN	;FUN=5D #29 reserved
	DW	BADFUN	;FUN=5E #30
	DW	BADFUN	;FUN=5F #31
	DW	BADFUN	;FUN=60 #32 reserved
	DW	CHKEY	;FUN=61 #33 get keyboard status
	DW	BADFUN	;FUN=62 #34 reserved
	DW	GETREG	;FUN=63 #35 CPU register array address
	DW	BADFUN	;FUN=64 #36
	DW	BADFUN	;FUN=65 #37
	DW	BADFUN	;FUN=66 #38
	DW	BADFUN	;FUN=67 #39
	DW	VCLEAR	;FUN=68 #40 clear screen
	DW	CLREOL	;FUN=69 #41 clear end of line
	DW	BADFUN	;FUN=6A #42
	DW	BADFUN	;FUN=6B #43
	DW	BADFUN	;FUN=6C #44
	DW	BADFUN	;FUN=6D #45
	DW	TEXT0	;FUN=6E #46 write zero-terminated string
	DW	SETFN	;FUN=6F #47 set filename
	DW	DELAY	;FUN=70 #48 delay mS
	DW	VNORM	;FUN=71 #49 normal video
	DW	VBOLD	;FUN=72 #50 bold video
	DW	INSLIN	;FUN=73 #51 insert line
	DW	DELLIN	;FUN=74 #52 delete line
	DW	CRTINI	;FUN=75 #53 crt init
	DW	CRTEXI	;FUN=76 #54 crt exit
	DW	RESTOR	;FUN=77 #55 reset disk system
	DW	SYSVEC	;FUN=78 #56 get system vector
	DW	HEXBYT	;FUN=79 #57 print hex byte
	DW	SETFCB	;FUN=7A #58 parse CP/M FCB
	DW	BADFUN	;FUN=7B #59
	DW	BADFUN	;FUN=7C #60
 	DW	MLCALL	;FUN=7D #61 Direct machine language call function
	DW	OUTB	;FUN=7E #62 8080/Z-80 Port Output function
	DW	INBYT	;FUN=7F #63 8080/Z-80 Port Input function

;.  The DISPLY table holds the base addresses (on the heap) of
;.  the variables for a given dynamic nesting level.
;.

DISPLY:	DW	0	;.LEVEL 0
	DW	0	;.LEVEL 1
	DW	0	;.LEVEL 2
	DW	0	;.LEVEL 3
	DW	0	;.LEVEL 4
	DW	0	;.LEVEL 5
	DW	0	;.LEVEL 6
	DW	0	;.LEVEL 7

PC:	DW	0	;.temporary place to save I2L PC
HP:	DW	0	;.holds the I2L heap pointer

LEVEL:	DW	0	;holds the current nesting level.

RBFR:	DS	80H	;CP/M read buffer
WBFR:	DS	80H	;CP/M write buffer

RFCB:	DS	40	;CP/M read File Control Block
WFCB:	DS	40	;CP/M write File Control Block

RPTR:	DW	RBFR	;CP/M read buffer pointer
WPTR:	DW	WBFR	;CP/M write buffer pointer

REMAIN:	DW	0	;remainder for arith
SIGN:	DB	0	;sign flag, arith
SFLG:	DB	0	;another arith flag
REG1:	DW	0	;temp arith register
REG2:	DW	0	;another temp arith register
CHAR:	DB	0	;single character buffer
INPFLG:	DB	0	;indication of input number
INBUPT:	DW	0	;input buffer pointer
SEED:	DW	1,0	;for random number generator

DEVNO:	DW	0	;.device number of current active device
			;reserved an extra byte so one can use
			;the POP H, SHLD DEVNO sequence, which
			;is much more efficient than POP H,
			;MOV A,L, STA DEVNO (this trick is used
			;in CHIN, CHOUT, etc.)

;  I/O related stuff

PROMPT:	DB	0;':'	;console input prompt character
			;  if 0 no prompt char will be issued
BADDEV:	DB	0	;place to save bad device number
ACPARM:	DB	0	;IO parameter
FUNC:	DB	0	;IO function number

;  For the loader

AUTO:	DB	0	;autorun
LODDEV:	DB	0	;.device number of the LOAD device
BASADR:	DW	LOSPAC	;.holds base addr to start loading at
TEMP:	DW	0	;.temp used to hold relocation constant
ADRES:	DW	0	;.holds addr read in from load device

HEAPHI:	DW	HISPAC	;holds highest available heap address
PROGLO:	DW	LOSPAC	;first address of program
STKLO:	DW	0	;first address of available heap
ECHFLG:	DB	1	;don't echo to console if nonzero
RUBCHR:	DB	RUB	;character to echo on rubout

RERUNF:	DB	0	;.if non-zero, indicates we had a good run

MONAD:	DW	MONITR	;.holds entry point to monitor or
			;.  operating system

;
;	KEYIN - CP/M buffered console input

KEYIN:	MVI	C,1	;function number to get a char from
			;  the console keyboard
	CALL	CPMIO	;get the char
	ORA	A	;clear carry flag
	RET

;	CONINO - Console initialization (clear screen and home cursor)
;
;CONINO:	MVI	A,FFCHR	;send a form feed to clear the screen
;	JMP	TVOUT	;fall through to TVOUT

;	TVOUT - CP/M buffered console output
;			(Control P causes output to be echoed on printer)

TVOUT:	LXI	H,POS0	;increment the head position counter
	CPI	FFCHR	;formfeed?
	JZ	TVOUT1
	CALL	POSCTR
	MOV	E,A	;put char into reg E as needed by CP/M
	MVI	C,2	;function number for write console
	CALL	CPMIO
	ORA	A	;clear carry flag
	RET

TVOUT1:	MVI	M,0
	INX	H
	MVI	M,0
	JMP	CLS

;	PRTOUT - printer output routine

PRTOUT:	LXI	H,POS4	;calculate the new value of the head
	CALL	POSCTR	;  position counter
	MOV	E,A	;put char into reg E as needed by CP/M
	MVI	C,5	;function number for write printer
	CALL	CPMIO	;write the character
	ORA	A	;clear the carry flag
	RET

;	KBORD - unbuffered console input routine

KBORD:	MVI	A,3
	CALL	CBIOS
	ORA	A	;make sure carry flag is cleared
	RET

;	CRT - unbuffered console output routine (^P has no effect)

CRT:	LXI	H,POS6	;update head position counter
	CPI	FFCHR	;formfeed?
	JZ	TVOUT1
	CALL	POSCTR
CRT1:	MOV	C,A
	MVI	A,4
	CALL	CBIOS
CRTN:	ORA	A	;make sure carry flag is cleared
	RET

POSCTR:	CPI	' '	;is it a control code?
	JNC	POSC2	;..no
	CPI	0DH	;is it a carriage return?
	JNZ	POSC0	;..no
	MVI	M,0	;..yes, so zero position
	INX	H
	MVI	M,0
	RET
POSC0:	CPI	9	;is it a tab?
	RNZ		;..no, so no adjustment needed
	PUSH	PSW	;..yes, so move to next field of 8
	MOV	A,M
	ADI	8
	ANI	0F8H
	MOV	M,A
	POP	PSW
	RET
POSC2:	PUSH	PSW	;save reg A
	MOV	A,M	;increment head position counter
	INR	A
	MOV	M,A
	INX	H
	MOV	A,M
	ACI	0
	MOV	M,A
	POP	PSW	;restore reg A
	RET

;
;	Set CPU stack and heap top
;
;  CPU stack is placed in high memory and grows down
;  while the heap grows up. Top of heap is nominally
;  set 1024 bytes below the stack.
;

SETSTK:	POP	B
	LHLD	CPMIO+1	;get BDOS base
	MVI	L,0
	SPHL		;set up the stack pointer
	LXI	D,-400H
	DAD	D
	SHLD	HEAPHI
	PUSH	B
	RET

;
;	MODULE:  Body of I2L interpreter
;

START:	CALL	SETSTK	;reset stack pointer
	CALL	RSEED	;preset random number seed
	MVI	A,0FFH	;reset console input buffer
	STA	INBUPT
	LHLD	STKLO	;reset stack (heap) pointer
	SHLD	HP
	SHLD	DISPLY	;fix display vector
	SUB	A
	STA	LEVEL	;set present level to zero

; Copy the command tail into pseudo device 8

	CALL	PSOPNO	;(IN CASE OF RESTART)
	CALL	PSOPNI
	LXI	H,DDMA
	PUSH	H
	MOV	E,M	;get length
	INX	H
	MVI	D,0
	MVI	A,126	;max length
	CMP	E
	JNC	START1
	MOV	E,D	;illegal so zero it
START1:	DAD	D
	MOV	M,D	;insert null at end
	POP	B
START2:	INX	B
	LDAX	B
	CPI	0
	JZ	START3	;done
	CALL	PSOUT
	JMP	START2
;
;  Now we fake a call to the user program as a procedure
;  (which it is).  The return address for this procedure
;  is the first byte of the program which has been preset
;  to be an exit opcode.  (See LOSPAC.)
;

START3:	LXI	H,0
	PUSH	H	;dummy level 0 save (two bytes)
	PUSH	H	;dummy base pointer push (two bytes)
	LHLD	PROGLO	;save the PC for return from user
	PUSH	H	;  program
	INX	H	;move past the EXIT opcode at LOSPAC
	SHLD	PC
	MOV	B,H	;copy I2L PC into BC
	MOV	C,L
	JMP	OPGO	;enter dispatch to execute

;.
;.  Main opcode fetch and dispatch
;.
;.  Here to fetch the I2L opcode pointed to by the I2L program
;.  counter.  This routine checks the validity of the opcode,
;.  and dispatches to the proper routine to perform the function
;.  needed.
;.

OPGOI:	INX	B	;.here if the PC needs incrementing
OPGO:	LDAX	B	;.fetch the opcode via I2L PC
	INX	B	;.bump PC to point to next byte
	ADD	A	;.shift left to form offset
	JC	GLOB	;..magic instruction
	JM	OPERR	;..still must be < 128 (otherwise it was
			;  a machine language function and the
			;  I2L PC must be preserved first).
;
;  common routine for OPGO and CMLDO
;
DOOP:	MOV	E,A	;make it a 16 bit offset
	MVI	D,0
	LXI	H,OPTAB	;get address of table of addresses
	DAD	D	;add the offset
	MOV	E,M	;get the address of the desired function
	INX	H
	MOV	D,M
	XCHG
	PCHL		;.and go to it

;
;	GLOB -
;
;  Magic instruction for fast load of global integer which is
;  the most common I2L instruction.
;
;  INPUT-
;	One byte instruction with high bit set.  The other
;	seven bits is the number of words between the address
;	in DISPLY and the address of the word holding the
;	integer.
;	The level is assumed = 0.
;
;  OUTPUT-
;	The integer on TOS
;

GLOB:	MOV	E,A
	MVI	D,0
	LHLD	DISPLY ;we know level = 0
GLOB1:	DAD	D	;HL points to address
	MOV	E,M	;get low byte of integer
	INX	H
	MOV	D,M	;and the high byte
	PUSH	D	;put the integer on the stack
	JMP	OPGO	;and fetch the next I2L opcode

;$00
;	EXIT instruction processor - 1 bytes
;
;.  Here to execute the EXIT and ABORT I2L instructions.
;.  Return to the monitor (operating system, etc.).

EXITDO:	MVI	A,1	;.indicate we had a good run
	STA	RERUNF	;.  in global flag
;$50
ABORT:	LHLD	MONAD	;.get addr of routine to return to
	PCHL		;.go there

;$01
;.	LOD instruction processor - 4 bytes
;.
;.  Here to execute the I2L LOD instruction.  It fetches an
;.  integer variable's value and places it on top of the stack.
;.

LODDO:	CALL	VADDR	;get variable's addr into HL
LODD1:	MOV	E,M	;.get low order byte of variable
	INX	H	;.point to high order
	MOV	D,M	;.get it
	PUSH	D	;.place it on stack
	JMP	OPGO	;and go process next I2L instruction

;
;  VADDR - Get the variable's addr into HL
;
;  This routine is used by LOD, STO, INC, DEC, ADR
;
;  INPUT-
;	BC   points to 1-byte containing static level number
;	BC+1 points to 2-byte offset from base address of
;		that level to the word containing the variable
;
;  OUTPUT-
;	HL contains addr of the variable
;	BC has been incremented by 3
;	A, DE and flags trashed
;

VADDR:	LDAX	B	;get static level number of variable
	INX	B
	MOV	E,A	;make it a 16 bit value in DE
	MVI	D,0
	LXI	H,DISPLY ;base addr of variables
	DAD	D	;calculate addr of word holding base
			;  addr of variables at this level
	MOV	E,M	;get the base addr of variables at
	INX	H	;  this level
	MOV	D,M
	XCHG
	LDAX	B	;fetch 16 bit offset of variable
	INX	B
	MOV	E,A
	LDAX	B
	INX	B
	MOV	D,A
	DAD	D	;form actual address of variable in HL
	RET

;$02
;	LDX instruction processor - 1 bytes
;
;  Here to execute the LDX I2L opcode.  This instruction uses
;  the value in TOS as an offset and the address in NOS and
;  fetch the one byte value from memory (used for arrays).
;

LDXDO:	POP	H	;TOS
	POP	D	;NOS
	DAD	D	;add them
	MOV	E,M	;now fetch the byte pointed to
	MVI	D,0	;make 16 bit value with high order zero
	PUSH	D	;place on stack
	JMP	OPGO

;$03
;.	STO instruction processor - 4 bytes
;.
;.  Here to execute the I2L STO instruction.  It takes the value
;.  off the top of the stack and stores it as an integer variable's
;.  value.

STODO:	CALL	VADDR	;get variable's addr into HL
	POP	D	;.get the value from top of stack
	MOV	M,E	;.now store low order into memory
	INX	H	;.point to next byte
	MOV	M,D	;.store high order into memory
	JMP	OPGO	;and go process next I2L instruction

;$04
;	STX instruction processor - 1 bytes
;
;  Here to execute the STX I2L instruction.  This instruction
;  takes the low byte of TOS and stores it in the address in NOS.
;

STXDO:	POP	D	;get 16 bit value from stack
	POP	H	;address
	MOV	M,E	;store low order byte into memory
	JMP	OPGO	;and go fetch next I2L instruction

;$05
;.	CAL instruction processor
;.
;.  Here to execute the CAL I2L instruction.  This instruction
;.  is used to link to (CALL) an I2L procedure.  After a procedure
;.  call the hardware stack is set up as follows:
;
;	Old level number
;	Old display vector at new level
;	Return I2L PC
;	...Stack pointer now points here
;
;.  The display vector for the new level is set to the next
;.  available heap location.
;.

CALDO:	LHLD	LEVEL	;.save current level number
	PUSH	H	;.  on stack
	LDAX	B	;.get new level number (of the procedure
	INX	B	;.  we are calling) and bump PC
	MOV	L,A
	SHLD	LEVEL	;save as current level number
	LXI	D,DISPLY ;get base addr of display table
	DAD	D	;calc addr for current level
	MOV	E,M	;.get low order value
	INX	H	;.point to high order display vector value
	MOV	D,M	;.get it
	PUSH	D	;.save old display vector on stack
	XCHG		;.place pointer in DE
	LHLD	HP	;.get heap pointer into HL
	XCHG		;.heap pointer in DE, pointer to display
			;.  vector in HL
	MOV	M,D	;.now make current heap pointer the base
	DCX	H	;.  address of new local variables
	MOV	M,E
	JMP	JSRDO

;$06
;.	RET instruction processor - 4 bytes
;.
;.  Here to execute the RET I2L instruction.  This instruction
;.  returns from an I2L procedure.  This pops the stuff pushed
;.  on the stack by the CAL.  It restores the heap to the depth
;.  it had at the time of the CAL.  The stack must have been
;.  returned to the value it had at the start of the procedure.
;.  This is why an incorrect number of arguments in a machine
;.  language function call will send I2L out to lunch.
;.  (The function will pop the wrong number of arguments and
;.  mess up the stack.)
;.

RETDO:	POP	B	;.get I2L PC to return to
	LXI	D,DISPLY ;get base addr of display table
	LHLD	LEVEL	;and current level
	DAD	D	;get addr of word for the current level
	MOV	E,M	;cut the heap back to what it was at
	INX	H	;  the time of the call
	MOV	D,M
	XCHG
	SHLD	HP
	XCHG
	POP	D	;get the old display vector again
	MOV	M,D	;.restore old display vector value
	DCX	H
	MOV	M,E
	POP	H	;.get old level number
	SHLD	LEVEL	;.save level number as current level
	JMP	OPGO	;.and go fetch I2L instruction after CAL

;$07
;.	JMP instruction processor - 3 bytes
;.
;.  Here to execute the unconditional JMP I2L instruction.  It
;.  loads a new PC value from the two bytes following the
;.  opcode.
;.

JMPDO:	LDAX	B	;.fetch low order of new PC value
	INX	B	;.point to high order of new PC
	MOV	E,A	;.save low order
	LDAX	B	;.get high order
	MOV	B,A	;.now place new PC into BC pair
	MOV	C,E
	JMP	OPGO	;.and go fetch next I2L instruction via
			;.  new PC

;$08
;.	JPC instruction processor - 3 bytes
;.
;.  Here to execute the conditional JMP instruction.
;.  *** This instruction jumps on false (zero value) ***
;.  A true is any non-zero value.  The value is on the top of
;.  the stack.
;.

JPCDO:	POP	D	;.get the conditional value off the stack
	MOV	A,E
	ORA	D
	JZ	JMPDO	;...false, so perform the jump
	INX	B	;...true, so skip over address bytes
	JMP	OPGOI	;and fetch the next I2L instruction

;$09
;.	HPI instruction processor - 3 bytes
;.
;.  Here to execute the HPI I2L instruction.  This instruction
;.  increments the heap pointer by some amount.  It is used to
;.  reserve heap space for local variables and parameters at
;.  run time.
;.

HPIDO:	LDAX	B	;.get the number of bytes to skip over
	INX	B
	MOV	E,A	;.form 16 bit value in DE
	LDAX	B
	INX	B
	MOV	D,A
	LHLD	HP	;.get heap pointer in HL
	DAD	D	;.add in increment
	SHLD	HP	;.store back as current heap pointer
	JMP	OPGO	;and go fetch next I2L instruction

;$0A
;.	ARG instruction processor
;.
;.  Here to execute the ARG I2L instruction.  This instruction
;.  retrieves procedure arguments from the stack and puts them
;.  on the heap.  The called procedure will then reserve the
;.  space with an HPI and the variables thus created will then
;.  have been preset to the appropriate values.
;
;	ARG = number of bytes - 1  (always an odd number)
;.

ARGDO:	LDAX	B	;.get number of bytes to copy
	MOV	E,A	;.16 bit number of bytes in DE
	MVI	D,0
	LHLD	HP	;.get heap pointer into HL
	DAD	D	;.calculate place to put bytes into
	ORA	A	;.clear carry flag
	RAR		;.make count the number of words to move
ARGLOP:	POP	D	;.get a word from stack
	MOV	M,D	;.high order first on heap since we are
	DCX	H	;.  moving down in memory
	MOV	M,E
	DCX	H
	DCR	A	;.done moving them all?
	CPI	0FFH
	JNZ	ARGLOP	;...no, so keep moving words
	JMP	OPGOI	;and go fetch next I2L instruction

;$0B
;.	IMM instruction processor - 3 bytes
;.
;.  Here to execute the IMM I2L instruction.  This instruction
;.  takes a 16 bit constant, which immediately follows the
;.  opcode, and stores it on the top of stack.
;.

IMMDO:	LDAX	B	;.get low order byte of constant
	INX	B	;.bump PC
	MOV	L,A	;.save low order byte
	LDAX	B	;.get high order byte of constant
	MOV	H,A	;.get high order
	PUSH	H	;.place value on top of stack
	JMP	OPGOI	;and go fetch next I2L instruction

;$0C
;.	CML instruction processor - 2 bytes
;.
;.  Here to execute the CML I2L instruction.  This instruction
;.  calls a machine language function.  Note that machine language
;.  functions are not written as subroutines.  They return by a
;.  direct jump to CMLRET.  Arguments to functions will be on the
;.  stack in the order in which they appeared in the parameter
;.  list of the call.
;.
;.  The result of a function, if any, is returned on the stack.
;.
;.  Function codes are >= 64 but < 128.
;.

CMLDO:	LDAX	B	;fetch the opcode via I2L PC
	INX	B	;bump PC to point to next byte
	MOV	H,B	;update memory image of I2L PC since a
	MOV	L,C	;  machine language function is allowed
	SHLD	PC	;  to use reg BC
	ADD	A	;shift left to form offset
	JC	BADFUN	;..jumps if MSB was on, opcode in error
	JM	DOOP	;..opcode OK, so go do it
	JMP	BADFUN	;opcode was < 64, so it was
			;  not a machine language function and
			;  we came to the wrong place.

;.  All machine code routines return here when done

CMLRET:	LHLD	PC	;.get back the I2L PC
	MOV	B,H
	MOV	C,L
	JMP	OPGO	;.and go fetch next I2L instruction

;$0D
;.	Arithmetic opcode processors
;.
;.	ADD instruction processor
;.
;.	    TOS := TOS + NOS
;.
;.  Here to execute the ADD I2L instruction.  This instruction
;.  adds the top of the stack to the next on the stack and
;.  places the result back on the stack.
;.

ADDDO:	POP	D	;.get the top of stack value
	POP	H	;.get next of stack value
	DAD	D	;.add them
	PUSH	H	;.place result back on stack
	JMP	OPGO	;and go fetch next I2L instruction

;$0E
;.	SUB instruction processor
;.
;.	    TOS := NOS - TOS
;.

SUBDO:	POP	D	;.get TOS value
	POP	H	;.get NOS value
	MOV	A,L	;.get low order NOS
	SUB	E	;.NOS - TOS low orders
	MOV	L,A	;.save low order difference in L
	MOV	A,H	;.get NOS high order
	SBB	D	;.NOS - TOS high orders
	MOV	H,A	;.save high order difference in H
	PUSH	H	;.put difference back on stack
	JMP	OPGO	;and go fetch next I2L instruction

;$0F
;.	MUY instruction processor
;.
;.	    TOS := TOS * NOS
;.

MUYDO:	POP	D	;.get TOS value
	POP	H	;.get NOS value
	CALL	MULT	;.do the multiply
	PUSH	D	;.put result on the stack
	JMP	OPGO	;and go fetch next I2L instruction

;$10
;.	DIV instruction processor
;.
;.	    TOS := NOS/TOS
;.

DIVDO:	POP	D	;get TOS value
	POP	H	;get NOS value
	CALL	DIVIDE	;.do the divide
	LHLD	REG1	;get the quotient
	PUSH	H	;put result on the stack
	JMP	OPGO	;and go fetch next I2L instruction

;.
;.	Integer math routines
;.
;.  These routines MUST prserve BC, which is the register pair
;.  used as the I2L program counter.
;.

;	16*16 multiply
;	result: 32 bit unsigned (HLDE) or 16 bit signed (DE)

;	HLDE <- HL * DE

MULT:	PUSH	B
	MOV	B,H
	MOV	A,L
	CALL	MPYX
	PUSH	H
	MOV	H,A
	MOV	A,B
	MOV	B,H
	CALL	MPYX
	POP	D
	MOV	C,D
	DAD	B
	ACI	0
	MOV	D,L
	MOV	L,H
	MOV	H,A
	POP	B
	RET

;	AHL <- A * DE

MPYX:	LXI	H,0
	MVI	C,8
MPYX1:	DAD	H
	RAL
	JNC	MPYX2
	DAD	D
	ACI	0
MPYX2:	DCR	C
	JNZ	MPYX1
	RET

;

HCHK:	MOV	A,H	;look at sign of HL
	ORA	A	;set flags
	RP
	INR	B	;diddle sign flag register

;	negate HL

NEGG:	SUB	A	;form 2's complement of HL
	SUB	L
	MOV	L,A
	MVI	A,0	;zero reg A without disturbing carry
	SBB	H
	MOV	H,A
	RET

;

DIVIDE:	MOV	A,D	;see if divisor is zero
	ORA	E
	JZ	DIVZ	;..yes, so show error
	PUSH	B	;preserve BC at all costs since it is
			;  the I2L program counter
	MVI	B,0	;preset sign flag register
	CALL	HCHK	;check sign of dividend and make it pos
	SHLD	REMAIN	;save it as the interim remainder
	XCHG
	CALL	HCHK	;check sign of divisor and make it pos
	SHLD	REG2	;save it
	CALL	DIV1	;do the division (pos number/pos number)
	DCR	B	;see if sign of result to be changed
	POP	B	;restore I2L PC
	RNZ		;..sign OK, so return
	LHLD	REMAIN	;get remainder
	CALL	NEGG	;negate it
	SHLD	REMAIN
	LHLD	REG1	;get quotient
	CALL	NEGG	;negate it
	SHLD	REG1
	RET

DIV1:	PUSH	B	;preserve sign flag register
	LXI	H,0	;initialize quotient to zero
	SHLD	REG1
	MVI	B,0	;initialize left shift counter
			;  (Contains the number of left shifts
			;  needed to normalize the divisor.)
;
;  Normalize the divisor (left shift until a 1 appears in MSB)
;
D1:	LHLD	REG2	;get the divisor
D1A:	SUB	A	;clear carry flag
	ORA	H
	JM	D1B	;..normalized
	INR	B	;update the left shift counter
	DAD	H	;left shift divisor (by multiplying by 2)
	JMP	D1A	;keep doing it until divisor normalized

D1B:	SHLD	REG2	;preserve normalized divisor
D2:	LHLD	REG1	;get the interim quotient
	DAD	H	;multiply by 2
	SHLD	REG1	;and save it
	LHLD	REG2	;get divisor
	SUB	A	;clear carry (otherwise right shift gets
			;  messed up)
	MOV	A,H	;divide quotient by 2 (right shift)
	RAR
	MOV	H,A
	MOV	A,L
	RAR
	MOV	L,A
	SHLD	REG2	;save divisor
	CALL	NEGG	;negate divisor
	XCHG
	LHLD	REMAIN	;get interim remainder
	DAD	D	;HL = interim remainder - normalized divisor
	JNC	D4	;..result was negative, so do not change
			;  REMAIN and REG1 (quotient)
	SHLD	REMAIN	;..result was positive, so save new
			;  interim remainder
	LHLD	REG1	;and add 1 to interim quotient
	INX	H
	SHLD	REG1
D4:	DCR	B	;decrement shift counter
	JNZ	D2	;..norm divisor >= divisor
	POP	B	;restore sign flag register
	RET

DIVZ:	MVI	A,DIVERR ;get code for divide by zero error
	JMP	ERROR	;and report error

;$11
;.	NEG instruction processor
;.
;.	    TOS := -TOS
;.

NEGDO:	POP	H	;.get TOS value
	CALL	NEGG
	PUSH	H	;.put result on stack
	JMP	OPGO	;and fetch next I2L instruction

;$12
;.	Boolean opcode processors
;.
;.  All I2L boolean instructions perform a comparison of the
;.  top two values on the stack.  They return either a TRUE
;.  (all ones) or a FALSE (all zeroes) value to the top of the
;.  stack.
;.
;.
;.	EQ instruction processor
;.
;.	TOS := TRUE if NOS = TOS, FALSE otherwise
;.

EQDO:	POP	D	;.get TOS value
	POP	H	;.get NOS value
	MOV	A,L	;.get low order NOS
	CMP	E	;.are low orders equal?
	JNZ	FALSE	;...no
	MOV	A,H	;...yes, so test high orders
	CMP	D
	JNZ	FALSE	;...high orders not equal
TRUE:	LXI	D,0FFFFH
	PUSH	D	;.put a TRUE on the stack
	JMP	OPGO	;and fetch next I2L opcode

;$13
;.	NE instruction processor
;.
;.	TOS := TRUE if NOS >< TOS, FALSE otherwise
;.

NEDO:	POP	D	;.get TOS value
	POP	H	;.get NOS value
	MOV	A,L	;.get low order NOS
	CMP	E	;.are they equal?
	JNZ	TRUE	;...no
	MOV	A,H	;.get high order NOS
	CMP	D
	JNZ	TRUE	;...not equal
FALSE:	LXI	D,0
	PUSH	D	;.put FALSE on stack
	JMP	OPGO	;and fetch next I2L opcode

;

CMPU:	MOV	A,H	; unsigned compare - C if HL < DE
	CMP	D	;                  - Z if HL = DE
	RNZ
	MOV	A,L
	CMP	E
	RET

CMPS:	MOV	A,H	; signed compare - C if HL < DE
	XRA	D
	JP	CMPU
	MOV	A,H
	ORA	A
	RP
	STC
	RET

;$14
;.	GE instruction processor
;.
;.	TOS := TRUE if NOS >= TOS, FALSE otherwise
;.

GEDO:	POP	D	;.get TOS value
	POP	H	;.get NOS value
	CALL	CMPS
	JNC	TRUE	;..NOS >= TOS
	JMP	FALSE	;..NOS < TOS

;$15
;.	GT instruction processor
;.
;.	TOS := TRUE if NOS > TOS, FALSE otherwise
;.

GTDO:	POP	H	;.get TOS value
	POP	D	;.get NOS value
	CALL	CMPS
	JC	TRUE	;..NOS > TOS
	JMP	FALSE	;..NOS <= TOS

;$16
;.	LE instruction processor
;.
;.	TOS := TRUE if NOS <= TOS, FALSE otherwise
;.

LEDO:	POP	H	;.get TOS value
	POP	D	;.get NOS value
	CALL	CMPS
	JNC	TRUE	;..NOS <= TOS
	JMP	FALSE	;..NOS > TOS

;$17
;.	LT instruction processor
;.
;.	TOS := TRUE if NOS < TOS, FALSE otherwise
;.

LTDO:	POP	D	;.get TOS value
	POP	H	;.get NOS value
	CALL	CMPS
	JC	TRUE	;..NOS < TOS
	JMP	FALSE	;..NOS >= TOS

;$18
;.	FOR instruction processor
;.
;.  Here to execute the FOR I2L instruction.  This instruction
;.  handles the test and branch of the FOR loop.  The stack
;.  is assumed to contain the limit and a copy of the loop
;.  variable (in that order).  They are compared and if the
;.  loop variable is > the limit, then the opcode argument is
;.  branched to and the stack is cleaned up, otherwise the I2L
;.  PC is advanced to the next opcode and the limit value left
;.  on the stack.

FORDO:	POP	D	;.get loop counter value
	POP	H	;.get upper limit value
	CALL	CMPS	;perform signed compare
	JC	JMPDO	;exit loop if loop counter > upper limit
FORCNT:	PUSH	H	;.place upper limit back on stack
	INX	B	;skip past 2 byte argument
	JMP	OPGOI	;and go fetch next I2L instruction

;$19
;.	INC instruction processor - 4 bytes
;.
;.  Here to execute the INC I2L instruction.  This instruction
;.  increments a variable's value and leaves a copy of the
;.  incremented value on the top of stack.  This is for use in
;.  loop constructs such as the FOR loop.  The stacked value must
;.  be correctly disposed of later, of course.
;.

INCDO:	CALL	VADDR	;get the variable's addr into HL
	MOV	E,M	;.get low order
	INX	H	;.point to high order
	MOV	D,M	;.get it
	INX	D	;.increment the value
	MOV	M,D	;.store updated value back
	DCX	H	;.point to low order again
	MOV	M,E	;.store it back
	PUSH	D	;.also place incremented value on stack
	JMP	OPGO	;and go fetch next I2L instruction

;$1A
;.	OR instruction processor
;.
;.	TOS := bitwise OR of TOS, NOS
;.

ORDO:	POP	D	;.get TOS value
	POP	H	;.get NOS value
	MOV	A,H	;.get high order NOS
	ORA	D
	MOV	H,A
	MOV	A,L	;.now do low order
	ORA	E
	MOV	L,A
	PUSH	H	;.put result on the stack
	JMP	OPGO	;and fetch next I2L opcode

;$1B
;.	AND instruction processor
;.
;.	TOS := bitwise AND of TOS, NOS
;.

ANDDO:	POP	D	;.get TOS value
	POP	H	;.get NOS value
	MOV	A,H	;.get high order NOS
	ANA	D
	MOV	H,A
	MOV	A,L	;.now do low order
	ANA	E
	MOV	L,A
	PUSH	H	;.put result on the stack
	JMP	OPGO	;and fetch next I2L opcode

;$1C
;.	NOT instruction processor
;.
;.	TOS := bitwise NOT of TOS
;.

NOTDO:	POP	D	;.get TOS value
	MOV	A,D	;.get high order NOS
	CMA
	MOV	D,A
	MOV	A,E	;.now do low order
	CMA
	MOV	E,A
	PUSH	D	;.put result on the stack
	JMP	OPGO	;and fetch next I2L opcode

;$1D
;	DUPCAT instruction processor
;
;  This routinte adds another entry to the stack which is a
;  copy of what was on top of stack (i.e., TOS is duplicated).
;
;
;DUPCAT:	POP	H	;get TOS
;	PUSH	H	;put it back
;	PUSH	H	;copy it
;	JMP	OPGO	;and fetch the next I2L opcode

;$1D
;	XOR instruction processor
;
;	TOS := bitwise XOR of TOS, NOS
;

XORDO:	POP	D	;.get TOS value
	POP	H	;.get NOS value
	MOV	A,H	;.get high order NOS
	XRA	D
	MOV	H,A
	MOV	A,L	;.now do low order
	XRA	E
	MOV	L,A
XORDO1:	PUSH	H	;.put result on the stack
	JMP	OPGO	;and fetch next I2L opcode

;$1E
;	DBA instruction processor
;
;	    TOS := 2*TOS + NOS
;

DBADO:	POP	H	;get TOS
	DAD	H	;double TOS
	POP	D	;get NOS
	DAD	D	;HL=2*TOS + NOS
	PUSH	H	;put result back on stack
	JMP	OPGO	;and fetch the next I2L opcode

;$1F
;	STD instruction processor
;
;  This routine pops both TOS and NOS off the stack.  It then
;  stores the value TOS at address NOS.
;

STDDO:	POP	D	;get TOS (16 bit value)
	POP	H	;get NOS (16 bit address)
	MOV	M,E	;store low byte
	INX	H	;point to high byte of word
	MOV	M,D	;store high byte
	JMP	OPGO	;and fetch the next I2L opcode

;$20
;	DBX instruction processor
;
;  This routine forms 2*TOS+NOS and then uses that value as
;  the address of a word to push onto TOS.  Both the
;  old TOS and NOS are popped off the stack.
;

DBXDO:	POP	H	;get TOS
	DAD	H	;HL = 2*TOS
	POP	D	;get NOS
	DAD	D	;HL = 2*TOS+NOS
	MOV	E,M	;get low byte of word pointed to by HL
	INX	H
	MOV	D,M	;get high byte of word
	PUSH	D	;put it on the stack
	JMP	OPGO	;and fetch the next I2L opcode

;$21
;	ADR instruction processor - 4 bytes
;
;  Load the address of a variable onto the stack.
;  Conventional three byte instruction.
;

ADRDO:	CALL	VADDR	;get variable's addr into HL
	PUSH	H	;put addr on TOS
	JMP	OPGO	;and fetch the next I2L opcode

;$22
;	LDI instruction processor
;
;  This routine replaces TOS with the integer it points to.
;  One byte instruction.
;

LDIDO:	POP	H	;get TOS
	MOV	E,M	;get low byte of integer
	INX	H	;point to high byte
	MOV	D,M	;get it
	PUSH	D	;put integer on TOS
	JMP	OPGO	;and fetch the next I2L opcode

;$23
;	LDA instruction processor
;
;  This instruction gets an integer from an absolute address
;  and puts it on the top of the stack.
;  Three byte instruction:
;	OPCODE
;	LOW ADDRESS BYTE
;	HIGH ADDRESS BYTE
;

LDADO:	LDAX	B	;get low byte of addr
	MOV	L,A
	INX	B	;point to high addr byte
	LDAX	B
	MOV	H,A	;HL = address
	MOV	E,M	;get low byte of integer
	INX	H
	MOV	D,M	;DE = integer
	PUSH	D	;put integer on TOS
	JMP	OPGOI	;and fetch the next I2L opcode

;$24
;	SIMM instruction processor
;
;  Short Immediate Load of 8 bit signed value
;  This instruction extends the value to 16 bits and pushes
;  the 16 bit result onto the stack
;  Two byte instruction:
;	OPCODE
;	VALUE TO LOAD
;

SIMMDO:	LDAX	B	;get the value
	MOV	E,A	;put 8 bit value in E
	MVI	D,0	;clear high byte
	ORA	A	;set flags
	JP	SIMM1	;..positive
	DCR	D	;..negative, so set D = FF
SIMM1:	PUSH	D	;put value on TOS
	JMP	OPGOI	;and fetch the next I2L opcode

;$25
;	CAJMP instruction processor
;
;  Case Jump
;  It pops TOS, compares tos to NOS and takes the branch
;  only if NOS was not equal to TOS.

CAJMP:	POP	D	;get TOS
	POP	H	;get NOS
	PUSH	H
	MOV	A,E
	CMP	L
	JNZ	JMPDO
	MOV	A,D
	CMP	H
	JNZ	JMPDO
	INX	B	;advance PC
	JMP	OPGOI

;$26
;	JSR instruction processor
;
; OPTIMIZED PROCEDURE CALL. USED ONLY IF NO LOCAL VARIABLES ARE
; PRESENT.  SCOPE IS UNCHANGED THUS IT IS EQUIVALENT TO A MACHINE
; LANGUAGE JSR.

JSRDO:	LDAX	B	;.get low order address of procedure
	INX	B
	MOV	E,A
	LDAX	B	;.get high order addr of procedure to call
	INX	B
	MOV	D,A	;.addr of procedure to call is in DE pair
	PUSH	B	;.save I2L PC to return to when procedure
			;.  'RET's
	MOV	B,D	;.now make I2L PC point to procedure
	MOV	C,E
	JMP	OPGO	;.and go fetch first I2L instruction of
			;.  the procedure called

;$27
;	RTS instruction processor
;
;  OPTIMIZED RETURN, TO MATCH JSRDO

RTSDO:	POP	B	;restore the I2L PC
	JMP	OPGO

;$28
;	DRP instruction processor
;
;ROUTINE TO POP TOS & DISCARD IT

DRPDO:	POP	H		;POP STACK
	JMP	OPGO

;$29
;	EXT instruction processor
;
;CALL EXTERNAL SUBROUTINE.  THIS ALLOWS ACCESS TO APPLICATION-
;SPECIFIC, MACHINE-LANGUAGE SUBROUTINES. THE ABSOLUTE ADDRESS
;APPEARS IN THE INSTRUCTION DIRECTLY.  ARGUMENTS, IF ANY, ARE
;PASSED VIA THE STACK.

EXTDO:	LXI	H,CMLRET
	PUSH	H
	MOV	L,C
	MOV	H,B
	MOV	E,M
	INX	H
	MOV	D,M
	INX	H
	SHLD	PC
	XCHG
	PCHL

;$2A
;	FORD instruction processor
;
;  This instruction handles the test and branch of the FOR DOWNTO
;  loop.  The stack is assumed to contain the limit and a copy of
;  the loop variable (in that order).  They are compared and if the
;  loop variable is < the limit, then the opcode argument is
;  branched to and the stack is cleaned up, otherwise the I2L
;  PC is advanced to the next opcode and the limit value left
;  on the stack.

FORDDO:	POP	H	;get loop counter value
	POP	D	;get limit value
	CALL	CMPS	;perform signed compare
	JC	JMPDO	;exit loop if done
	PUSH	D	;place limit back on stack
	INX	B	;skip past 2 byte argument
	JMP	OPGOI	;and go fetch next I2L instruction

;$2B
;	DEC instruction processor - 4 bytes
;
;  Here to execute the DEC I2L instruction.  This instruction
;  decrements a variable's value and leaves a copy of the
;  decremented value on the top of stack.  This is for use in
;  loop constructs such as the FOR loop.  The stacked value must
;  be correctly disposed of later, of course.
;

DECDO:	CALL	VADDR	;get the variable's addr into HL
	MOV	E,M	;get low order
	INX	H	;point to high order
	MOV	D,M	;get it
	DCX	D	;decrement the value
	MOV	M,D	;store updated value back
	DCX	H	;point to low order again
	MOV	M,E	;store it back
	PUSH	D	;also place decremented value on stack
	JMP	OPGO	;and go fetch next I2L instruction
;

SAR:	MOV	A,H	; SHIFT ARITHMETIC RIGHT
	RLC
	RRC
SAR1:	RAR
	MOV	H,A
	MOV	A,L
	RAR
	MOV	L,A
	RET

; $3C
;	ARY instruction processor
;
; ROUTINE TO RESERVE A MULTI-DIMENSIONAL ARRAY
; FOR EXAMPLE: MKARRAY((3, 5, 7, 11), 4, 2, addr ARRAYNAME)\
; THE STACK CONTAINS:
;   THE NUMBER OF ELEMENTS IN EACH DIMENSION (3, 5, 7, 11)
;   THE NUMBER OF DIMENSIONS (4)
;   THE NUMBER OF BYTES IN EACH ELEMENT OF THE ARRAY (1=CHAR, 2=INT)
;   THE ADDRESS OF THE POINTER TO THE ARRAY (A POINTER TO A POINTER)
;    (NOTE: THIS ROUTINE SETS THE POINTER TO THE SPACE IT RESERVES)

; *** This code is preliminary and is adapted from the PC version. ***
; *** It has only been tested outside the I2L environment.         ***

MKARY:	POP	H	;GET ADDRESS OF POINTER TO ARRAY
	SHLD	APTR
	POP	H	;GET NUMBER OF BYTES IN EACH ELEMENT
	SHLD	SIZ
	POP	H	;GET NUMBER OF DIMENSIONS
	DCX	H	;CONVERT TO MAXIMUM INDEX OFF OF BP (LEVMAX)
	DAD	H	;DOUBLE FOR WORD ENTRIES ON THE STACK
	MOV	E,L
	MOV	D,H
	DAD	SP	;POINT TO NUMBER OF ELEMENTS IN 1ST DIMENSION
	SHLD	SPTR
	XCHG
	CALL	NEGG
	SHLD	LEVMAX
	LXI	H,0	;SET INDEX TO POINT TO SIZE OF FIRST DIMENSION
	SHLD	IDX
	PUSH	B
	CALL	GENARY	;RESERVE FIRST DIMENSION OF ARRAY AND STORE ITS
			;BASE ADDRESS INTO LOCATION POINTED TO BY BX
	POP	B
	LHLD	SPTR	;CLEAN UP STACK--DISCARD ARGUMENTS
	INX	H
	INX	H
	SPHL
	JMP	OPGO	;GO GET NEXT CODE

; RECURSIVELY CALLED SUBROUTINE FOR SETTING UP A MULTI-DIMENSIONAL ARRAY.
; RESERVE AN ARRAY AND STORE ITS BASE ADDRESS INTO THE LOCATION POINTED TO
;  BY THE BX REGISTER.
; REGISTER USAGE (* INDICATES INPUT VALUES):
;   *   APTR  - ADDRESS OF POINTER TO NEXT DIMENSION OF ARRAY
;       BC    - LOOP COUNTER AND INDEX
;       BADR  - BASE ADDRESS OF CURRENT DIMENSION OF ARRAY
;   *   SIZ   - NUMBER OF BYTES IN AN ELEMENT (1 OR 2)
;   *   IDX   - INDEX TO SIZE OF EACH DIMENSION, ALSO LEVEL OF RECURSION (* -2)
;   *   SPTR  - POINTS TO FIRST ARGUMENT ON THE STACK (SIZE OF FIRST DIMENSION)

GENARY:	LHLD	SPTR	;GET NUMBER OF ELEMENTS FOR CURRENT DIMENSION
	XCHG
	LHLD	IDX
	DAD	D
	MOV	E,M
	INX	H
	MOV	D,M

	MOV	C,E	;SAVE FOR POSSIBLE USE BY LOOP COUNTER BELOW
	MOV	B,D	;RESERVE SPACE ON HEAP
	PUSH	B

	LHLD	IDX	;IS THIS THE LAST DIMENSION?
	XCHG
	LHLD	LEVMAX
	CALL	CMPS
	JC	GA02	;SKIP IF NOT

	LDA	SIZ	;IS THIS A CHAR ARRAY?
	CPI	1
	JZ	GA07	;JUMP IF SO (LAST DIMENSION IS A SINGLE BYTE)
GA02:	POP	H
	DAD	H	;MULTIPLY ELEMENTS BY 2 BYTES PER INTEGER
	JC	GA10	;CHECK FOR MEMORY OVERFLOW: ERROR IF > $FFFF
	PUSH	H
GA07:	POP	H
	XCHG		;DX = BASE OF RESERVED HEAP SPACE
	LHLD	HP
	SHLD	BADR
	DAD	D
	SHLD	HP	;ADD NUMBER OF BYTES TO RESERVE TO HEAP POINTER
	JC	GA10	;CHECK FOR MEMORY OVERFLOW: ERROR IF > $FFFF
	XCHG
	LHLD	HEAPHI	;CHECK FOR MEMORY OVERFLOW
	XCHG
	CALL	CMPU
	JC	GA20	;SKIP IF OK

GA10:	MVI	A,RMERR	;MEMORY OVERFLOW
	JMP	ERRORP

GA20:	LHLD	BADR	;STORE BASE ADDRESS OF RESERVED SPACE INTO PTR
	XCHG
	LHLD	APTR
	MOV	M,E
	INX	H
	MOV	M,D

	LHLD	LEVMAX	;IS THIS THE LAST DIMENSION?
	XCHG
	LHLD	IDX
	CALL	CMPS
	RC		;JUMP IF SO--DON'T RECURSE ANY LOWER
	CALL	CMPU
	RZ		;JUMP IF SO--DON'T RECURSE ANY LOWER

	LXI	H,0	;INITIALIZE INDEX TO FIRST ELEMENT
GA30:	MOV	E,C
	MOV	D,B
	CALL	CMPU	;IS INDEX BEYOND LIMIT?
	RNC		;JUMP IF SO--EXIT LOOP

	PUSH	H	;SAVE INDEX
	DAD	H	;MULTIPLY BY 2 TO INDEX BY WORDS
	XCHG
	LHLD	BADR
	DAD	D
	SHLD	APTR	;ADD BASE ADDRESS OF RESERVED SPACE (CURRENT DIM)

	LHLD	IDX	;ADVANCE INDEX TO NEXT DIMENSION SIZE ON STACK
	DCX	H
	DCX	H
	SHLD	IDX

	PUSH	B	;SAVE SOME THINGS (LIMIT,BASE ADDR)
	LHLD	BADR
	PUSH	H

	CALL	GENARY	;RESERVE NEXT DIMENSION OF ARRAY AND STORE ITS

	POP	H	;BASE ADDRESS INTO LOCATION POINTED TO BY BX
	SHLD	BADR
	POP	B	;RESTORE THINGS

	LHLD	IDX	;RESTORE INDEX TO PREVIOUS DIMENSION SIZE ON STK
	INX	H
	INX	H
	SHLD	IDX

	POP	H
	INX	H	;NEXT ELEMENT
	JMP	GA30	;LOOP FOR EACH ELEMENT IN THIS DIMENSION

LEVMAX:	DS	2	;MAXIMUM LEVEL OF RECURSION (LAST DIMENSION)
SIZ:	DS	2
APTR:	DS	2
SPTR:	DS	2
BADR:	DS	2
IDX:	DS	2

;$3D
;	ASR instruction processor
;
; ARITHMETIC SHIFT RIGHT
;
; NOS ->> TOS
;

ASRDO:	POP	D	;GET SHIFT COUNT FROM TOS
	POP	H	;GET NOS
	INR	E
ASRDO1:	DCR	E
	JZ	XORDO1
	CALL	SAR
	JMP	ASRDO1

;$3E
;	LSL instruction processor
;
; LOGICAL SHIFT LEFT
; NOS << TOS
;

LSLDO:	POP	D	;GET SHIFT COUNT FROM TOS
	POP	H	;GET NOS
	INR	E
LSLDO1:	DCR	E
	JZ	XORDO1
	DAD	H
	JMP	LSLDO1

;$3F
;	LSR instruction processor
;
; LOGICAL SHIFT RIGHT
; NOS >> TOS
;

LSRDO:	POP	D	;GET SHIFT COUNT FROM TOS
	POP	H	;GET NOS
	INR	E
LSRDO1:	DCR	E
	JZ	XORDO1
	ORA	A
	MOV	A,H
	CALL	SAR1
	JMP	LSRDO1

;.
;.  Here when an I2L opcode is out of range or unimplemented.
;.

OPERR:	MVI	A,OPCERR ;.get error code number
	JMP	ERROR	;.and inform user

;$40 #00
;.	ABSFUN - Absolute Value function
;.
;.	    TOS := ABS(TOS)
;.

ABSFUN:	POP	H	;.get TOS value
	CALL	ABSA	;.take absolute value
	PUSH	H	;.put it on the stack
	JMP	OPGO	;and fetch the next I2L opcode

ABSA:	SUB	A	;zero reg A
	ADD	H	;get high byte of value
	RP		;..positive, so leave HL alone
	CMA		;..negative, so calc 2's complement
	MOV	H,A
	MOV	A,L
	CMA
	MOV	L,A
	INX	H	;changes 1's complement to 2's complement
	RET

;$41 #01
;	RANFUN
;
;  Return a random number between 0 and TOS-1
;

RANFUN:	POP	H
	PUSH	H
	MOV	A,L
	ORA	H
	JNZ	RANFU1
	CALL	RSEED
	JMP	CMLRET

RANFU1:	LHLD	SEED
	LXI	D,4E35H
	CALL	MULT
	PUSH	D
	PUSH	H
	LHLD	SEED+2
	LXI	D,4E35H
	CALL	MULT
	POP	H
	DAD	D
	PUSH	H
	LHLD	SEED
	LXI	D,015AH
	CALL	MULT
	POP	H
	DAD	D
	POP	D
	MVI	A,1
	ADD	E
	MOV	E,A
	MVI	A,0
	ADC	D
	MOV	D,A
	MVI	A,0
	ADC	L
	MOV	L,A
	MVI	A,0
	ADC	H
	MOV	H,A
	XCHG
	SHLD	SEED
	XCHG
	SHLD	SEED+2
	POP	D
	CALL	MULT
	PUSH	H
	JMP	CMLRET

RSEED:	LXI	H,2537	;reset seed
	SHLD	SEED
	LXI	H,5149
	SHLD	SEED+2
	RET

;$42 #02
;	REMFUN - return the remainder of the most recent division
;
;  INPUT-
;	Argument is an expression whose result is thrown away
;	The expression may contain one division, or be zero
;	to get the remainder of an earlier division.
;
;  OUTPUT-
;	Remainder on TOS
;

REMFUN:	POP	H	;throw away argument
	LHLD	REMAIN
	PUSH	H	;put the remainder on TOS
	JMP	OPGO	;and fetch the next I2L opcode

;$43 #03
;	RESERV - Reserve heap space
;
;  This is the trick by which XPL handles dynamic storage for
;  arrays.  Since the space is reserved in the heap allocation
;  of a procedure the array will disappear when the procedure
;  is exited.
;
;  INPUT-
;	TOS := number of bytes to reserve
;
;  OUTPUT-
;	TOS := address of beginning of reserved space
;

RESERV:	POP	D	;DE = number of bytes to reserve
	LHLD	HP	;get present heap pointer
	PUSH	H	;put on TOS as beginning addr of
			;  reserved space
	DAD	D	;calculate new heap pointer
	SHLD	HP	;update heap pointer
	SUB	A	;clear flags, zero reg A
	ADD	D	;add in D
	MVI	A,3	;set up error code for negative
			;  reservation
	JM	ERROR	;..negative reservation
	CALL	TESTER	;do we have enough space?
	JMP	OPGO	;and fetch the next I2L opcode

;$44 #04
;	SWAPFU - swap high and low bytes on TOS
;

SWAPFU:	POP	H	;get TOS
	MOV	D,L	;swap bytes
	MOV	E,H
	PUSH	D	;put result back on TOS
	JMP	OPGO	;fetch next opcode

;$45 #05
;	EXTEN - extend sign from low byte on TOS into high byte
;

EXTEN:	POP	H	;get TOS
	MOV	A,L
	RAL
	SBB	A
	MOV	H,A
	PUSH	H	;put result back on stack
	JMP	OPGO

;$46
;	SHORT - make TOS into sign byte + value byte
;
;  Make the TOS a 7 bit signed value + a sign byte.  This
;  makes it possible to store small signed integers in a
;  single array element.
;
;
;SHORT:	POP	H	;get TOS
;	SUB	A	;zero reg A, clear flags
;	ADD	H	;check high byte
;	JP	SHORT1	;..positive, so zero high byte
;	MOV	A,L	;..negative, so force low byte neg
;	ORI	80H
;	MOV	L,A
;SHORT1:	MVI	H,0
;	PUSH	H	;put result back on TOS
;	JMP	OPGO

;$46 #06
;	RESTRT - Restart the current program
;

RESTRT:	MVI	A,1	;indicate we had a good run
	STA	RERUNF
	JMP	START

;$47 #07
;	CHIN -
;  Get one ASCII char from device
;  INPUT-
;	Device number on TOS
;  OUTPUT-
;	ASCII char in low byte on TOS
;	00 in high byte on TOS
;

CHIN:	POP	H	;get device number off stack
	SHLD	DEVNO	;put it in device number byte in memory
	CALL	GETC	;get a char from that device in reg A
	MOV	L,A
	MVI	H,0	;zero the high byte of the word
	PUSH	H	;put char on stack
	JMP	OPGO	;and fetch the next I2L opcode

;$48 #08
;	CHOUT-
;  Output the character in low byte of word on top of stack
;
;  INPUT-
;	TOS := ASCII char
;	NOS := device number
;

CHOUT:	POP	D	;get char off stack
	POP	H	;get device number off stack
	SHLD	DEVNO	;put into device number byte in memory
	MOV	A,E	;put the char into reg A
	CALL	OUTTO	;output it to the device
	JMP	OPGO	;and fetch the next I2L opcode

;$49 #09
;	SKIP - Print End-of-Line
;
;  INPUT-
;	TOS := device number (in low order byte)
;

SKIP:	POP	H	;get device number from TOS
	SHLD	DEVNO
	CALL	CRLF	;output a cariage return-line feed pair
	JMP	OPGO	;and fetch the next I2L opcode

;
;	CRLF - Output a Carriage Return-Line feed pair
;
;  INPUT-
;	DEVNO contains the device number to get the CRLF
;

CRLF:	MVI	A,CR	;get the carriage return char
	CALL	OUTTO	;output it
	MVI	A,LF	;get the line feed char
	CALL	OUTTO	;output it
	RET

;$4A #10
;	INTIN - input a signed decimal integer to TOS
;
;  INPUT-
;	TOS := device number
;
;  OUTPUT-
;	TOS := signed integer
;

INTIN:	POP	H	;get device number
	SHLD	DEVNO	;set the device number
	CALL	GETNO	;get the signed decimal integer
	LHLD	REG1	;get the value from temp store
	PUSH	H	;and put it on TOS
	JMP	OPGO

;$4B #11
;	INTOUT - Output signed integer on TOS as decimal number
;
;  INPUT-
;	TOS := signed integer
;	NOS := device number
;

INTOUT:	POP	H	;get signed integer
	SHLD	REG1
	POP	H	;get device number
	SHLD	DEVNO
	CALL	INTO	;output integer in REG1 on device
	JMP	CMLRET

;$4C #12
;	TEXT -  Print a string
;
;  INPUT-
;	TOS := addr of beginning of string
;	NOS := device number (in low order byte)
;	String terminated by parity bit set on last character
;

TEXT:	POP	H	;get address of string
	POP	D	;get device number
	MOV	A,E
	STA	DEVNO
	CALL	TXTLOP	;output the string
	JMP	OPGO	;and fetch the next I2L opcode

;
;	TXTLOP - Output a string
;
;  INPUT-
;	DEVNO contains device number string is to be sent to
;	HL = address of string
;	String ended by char with parity bit set
;

TXTLOP:	MOV	A,M	;get the character
	INX	H	;point to next char
	ORA	A	;set flags
	JM	TXTL1	;..parity bit set, print last char
	CALL	OUTTO	;output character
	JMP	TXTLOP	;continue until find parity bit set

TXTL1:	ANI	7FH	;strip parity bit
	CALL	OUTTO	;output the character
	RET

;$4D #13
;	IDEVIN - Initialize input device
;
;  INPUT-
;	TOS := device number to be opened for input
;

IDEVIN:	POP	H	;get device number
	SHLD	DEVNO
	CALL	INIDEV	;go do it
	JMP	CMLRET

;$4E #14
;	ODEVIN - Initialize output device
;
;  INPUT-
;	TOS := device number to be opened for output
;

ODEVIN:	POP	H	;get device number
	SHLD	DEVNO
	CALL	INODEV	;go do it
	JMP	CMLRET

;$4F #15
;	ODEVCL - close output device
;
;  INPUT-
;	TOS := device number to be closed
;

ODEVCL:	POP	H	;get the device number
	SHLD	DEVNO
	CALL	CLODEV	;close the device
	JMP	CMLRET

;$51 #17
;	CALLER - Call an external procedure
;
;  This intrinsic will execute an "entire" procedure.  This
;  makes it possible for XPL to execute arrays as I2L code
;  (which they had better contain).  Since the procedure must
;  be "entire" (contain no external references) it can be only
;  called from level zero as a level zero procedure.  Used by
;  the resident version to execute the user's code.
;

CALLER:	POP	H	;get the new address
	SHLD	TEMP	;save it for later
	LHLD	LEVEL	;get current level
	SUB	A	;make sure it's level zero
	ORA	L
	JNZ	CALLE1	;..not level zero, so show error
	PUSH	H	;put the level on the stack
	LXI	H,DISPLY ;get the base addr of display table
;  level 0, so don't need to calculate offset due to level
	MOV	E,M	;get display vector into DE
	INX	H
	MOV	D,M
	PUSH	D	;put old display vector on the stack
	XCHG
	LHLD	HP	;get present heap location
	XCHG		;DE=heap pointer, HL=display table loc
	MOV	M,E	;make the current heap the base addr of
	DCX	H	;  new local variables
	MOV	M,D
	PUSH	B	;put old PC on the stack
	LHLD	TEMP	;get the new address again
	MOV	B,H	;and put into BC (I2L PC)
	MOV	C,L
	JMP	OPGO	;and fetch the I2L opcode at beginning
			;  of this new procedure

CALLE1:	MVI	A,CALERR ;caller at wrong level
	JMP	ERROR	;report error to user

;$52 #18
;	SPACE - Determine the remaining space
;
;  Intrinsic to tell a user how many bytes of space he has
;  left (on the heap).  Of course he may not reserve all of
;  it, since he must leave a working heap for subsequent
;  procedure calls, etc.  Since only the user knows how much
;  this might be it is left to hime to decide how much he
;  dares play with.
;
;  INPUT-
;	Values in HEAPHI and HP
;
;  OUTPUT
;	TOS := (contents of HEAPHI) - (contents of HP)
;

SPACE:	LHLD	HP
	CALL	TESTER	;calculate the difference between
			;  the contents of HEAPHI and HP
	PUSH	H	;put the result on the stack
	JMP	OPGO

;$53 #19
;	RERUN - Test the rerun flag
;
;  This intrinsic returns a TRUE on TOS if the current XPL
;  program has been run since the last call to the loader,
;  editor, etc.
;

RERUN:	LDA	RERUNF	;get the rerun flag
	MOV	L,A
	MVI	H,0
	PUSH	H	;and put it on TOS
	JMP	OPGO

;$54 #20
;	TSTSP - Get the current heap pointer
;
;  This intrinsic puts the current value of the heap pointer
;  on the TOS.  This function is used for saving the heap
;  pointer for later restoration with SETSP.  The user had
;  better have a good idea of the functioning of I2L before
;  dinging with the heap pointer or he will surely bomb himself.
;

TSTSP:	LHLD	HP	;get the heap pointer
	PUSH	H	;put it on TOS
	JMP	OPGO

;$55 #21
;	SETSP - Set heap pointer to value on TOS
;
;  This intrinsic sets the heap pointer -- a VERY DANGEROUS
;  thing to do.  See TSTSP's comments.
;

SETSP:	POP	H	;get value from TOS
	SHLD	HP	;and set heap pointer to it
	JMP	OPGO

;$56 #22
;	SETRUN - Set the rerun flag
;
;  Set the rerun flag to show that we have a good resident
;  program.
;

SETRUN:	POP	H
	MOV	A,L	;non-zero means OK
	STA	RERUNF
	JMP	OPGO

;$5A #26
;	HEXIN(DEVNO)
;
;  Get a hex number from device DEVNO.  The number is first initialized to
;  zero and then digits are collected.  The collection stops on the first
;  non-hex digit (ie, any char other than 0-9,A-F).  If more than 4 digits
;  are collected only the least significant 4 digits are kept.
;

HEXIN:	POP	H	;get device number
	SHLD	DEVNO
	PUSH	B	;save I2L program counter
	SUB	A
	STA	INPFLG
	LXI	H,0	;zero integer
HEXIN1:	PUSH	H
HEXIN2:	CALL	GETC
	CPI	'_'	;ignore any underlines
	JZ	HEXIN2
	CPI	EOFCHR
	JZ	HEXIN3
	CALL	INHEX3	;convert it to a value 0-15
	JNC	HEXIN4	;..valid hex digit
	LDA	INPFLG	;last digit?
	ORA	A
	JZ	HEXIN2	;try again
HEXIN3:	POP	H	;get the value of the integer
	POP	B	;restore I2L Program Counter
	PUSH	H	;put hex number on the stack
	JMP	OPGO

HEXIN4:	POP	H	;get previous value of the integer
	DAD	H	;left shift 4 (multiply by 16)
	DAD	H
	DAD	H
	DAD	H
	ADD	L	;add in current hex digit
	MOV	L,A
	MVI	A,1
	STA	INPFLG
	JMP	HEXIN1	;continue until get non-hex character

;$5B #27
;	HEXPRN(DEVNO,INTEGER)
;
;  Print the value in INTEGER as a hex number on device DEVNO.
;

HEXPRN:	POP	H	;get the integer
	POP	D	;get the device number
	MOV	A,E
	STA	DEVNO
	PUSH	B	;save the I2L PC
	CALL	HEXO	;print the value contained in HL as a hex number on
			;the specified device
	POP	B
	JMP	OPGO

;$61 #33
;	CHKEY - Get keyboard status
;
;  Intrinsic to test for key strike on the keyboard. Returns
;  true if key was struck. Use OpenI(1); to clear any pending
;  characters before or after ChKey.
;

CHKEY:	MVI	A,2
	CALL	CBIOS
	MOV	L,A
	MOV	H,A
	PUSH	H
	JMP	CMLRET

;$63 #35
;	GETREG - Get CPU register address
;
;  Get the address of the CPU register array returned by MLCALL
;

GETREG:	LXI	H,REGIM
	PUSH	H
	JMP	OPGO

;$6E #46
;	TEXT0 -  Print a zero-terminated string
;
;  INPUT-
;	TOS := addr of beginning of string
;	NOS := device number (in low order byte)
;	String ended by a null character
;

TEXT0:	POP	H	;get address of string
	POP	D	;get device number
	MOV	A,E
	STA	DEVNO
TXTL01:	MOV	A,M	;get the character
	INX	H	;point to next char
	ORA	A	;
	JZ	OPGO	;quit
	CALL	OUTTO	;output character
	JMP	TXTL01

;$6F #47
;	FNAME (STRING,MODE)
;
;  Set filename and mode for device 3. The filename STRING is in the
;  form X:FILENAME.EXT terminated by a null or bit 7 of the last
;  character set.  Leading spaces are skipped. X is an optional drive
;  letter; if omitted the currently logged drive will be used. If MODE
;  is zero the file is for input; non-zero indicates output.
;
;  INPUT-
;	TOS := 0=read, 1=write
;	NOS := addr of beginning of string
;
;	String ended by a null character or bit 7 set
;

SETFN:	POP	D	;get mode
	POP	H	;get address of string
	MOV	A,E
	LXI	D,RFCB
	ORA	A
	JZ	SETFN1
	LXI	D,WFCB
SETFN1:	CALL	MAKFCB
	JMP	OPGO

;$77 #55
;	RESTORE - Reset disk system
;

RESTOR:	PUSH	B
	MVI	C,GETDRV
	CALL	CPMIO
	PUSH	PSW
	MVI	C,RESETD
	CALL	CPMIO
	POP	PSW
	MOV	E,A
	MVI	C,LOGOND
	CALL	CPMIO
	POP	B
	JMP	OPGO

;$78 #56
;	SYSVEC - Get system vector
;

SYSVEC:	POP	H	;index
	DAD	H
	LXI	D,SVEC
	JMP	GLOB1

;$79 #57
;	HEXBYT(DEVNO,INTEGER)
;
;  Print the low order byte of the value in INTEGER as a hex
;  number on device DEVNO.

HEXBYT:	POP	H	;get the integer
	POP	D	;get the device number
	MOV	A,E
	STA	DEVNO
	PUSH	B	;save the I2L PC
	MOV	A,L	;get the low order byte
	CALL	HEX2	;print the low order byte as hex
	POP	B	;restore the I2L PC
	JMP	OPGO

;$7A #58
;	SETFCB(STRING,FCB)
;
;  Parse a filename into the 33 byte CP/M File Control Block at address
;  FCB. The filename STRING is in the form X:FILENAME.EXT terminated by
;  a null or bit 7 of the last character set. Leading spaces are skipped.
;  X is an optional drive letter; if omitted the currently logged drive
;  will be used. Wildcard characters '*' and '?' are permitted.
;
;  Note: Formerly called MAKEFCB. This function has been renamed to
;  reflect the changed order of arguments.

SETFCB:	POP	D	;get the addr of the FCB to be set up
	POP	H	;get the addr of the beginning of the string
	CALL	MAKFCB	;zero the FCB then fill in the drive number,
			;file name and extension
	JMP	OPGO	;get next I2L opcode and execute it

;$7D #61
;	MLCALL - direct machine language call
;
;  Call the routine whose address is the word at NOS with the
;  registers set to the values held in the register image area whose
;  address is TOS.  The contents of the registers are stored in REGIM
;  upon return from the routine.
;

MLCALL:	POP	H	;get address of image block
	POP	D	;get address of routine
	PUSH	B	;save BC (I2L program counter)
	LXI	B,MLRTN	;put return address on the stack
	PUSH	B
	PUSH	D	;put address of called routine on stack
	MOV	E,M	;get PSW image
	INX	H
	MOV	D,M
	INX	H
	PUSH	D	;and put it on the stack
	MOV	B,M
	INX	H
	MOV	C,M
	INX	H
	MOV	D,M
	INX	H
	MOV	E,M
	INX	H
	MOV	A,M
	INX	H
	MOV	L,M
	MOV	H,A
	POP	PSW	;all registers set up now
	RET		;go to machine language routine

MLRTN:	PUSH	H	;save reg HL
	LXI	H,REGIM	;point to beginning of register image block
	PUSH	D	;save DE
	PUSH	PSW
	POP	D	;get PSW into DE
	MOV	M,E	;save PSW in image area
	INX	H
	MOV	M,D
	INX	H
	MOV	M,B	;save BC in image area
	INX	H
	MOV	M,C
	INX	H
	POP	D
	MOV	M,D	;save DE in image area
	INX	H
	MOV	M,E
	INX	H
	POP	D
	MOV	M,D	;save HL in image area
	INX	H
	MOV	M,E
	POP	B	;restore BC (I2L program counter)
	JMP	OPGO	;get next I2L function

;$7E #62
;	OUTB - 8080/Z-80 Port Output
;
;  Output the low byte in NOS to the port whose number is the low byte TOS
;

OUTB:	POP	D	;get TOS
	POP	H	;get NOS
	MOV	A,L	;get port number
	STA	OUTB1	;set up output instruction
	MOV	A,E	;put value to be output in Reg A
	OUT	0
OUTB1	EQU	$-1
	JMP	OPGO	;get next I2L opcode and do it

;$7F #63
;	INBYT - 8080/Z-80 Port Input
;
;  Input a byte from the port whose number is the low byte TOS and put the
;  result as the low byte TOS (high byte is zeroed).

INBYT:	POP	H	;get word from TOS
	MOV	A,L	;get port number
	STA	INBYT1	;set up input instruction
	IN	0
INBYT1	EQU	$-1
	MOV	L,A	;put the value input as low byte
	MVI	H,0	;zero high byte
	PUSH	H	;  on the TOS
	JMP	OPGO	;get next I2L opcode and do it

;
;.  Here when a machine language function code is out of range
;.  or unimplemented.
;.

BADFUN:	MVI	A,ERRFUN ;.get error code number
			;.fall into error print routine

;
;.  Here when an error is encountered.  Error code number is
;.  in the A register.  Print a message on user's console.
;.

ERRORP:	LHLD	PC	;.get I2L PC since it was not
	MOV	B,H	;.  in the registers
	MOV	C,L
ERROR:	PUSH	B	;.save I2L PC at time of error
	PUSH	PSW	;.save error code number
	MVI	A,0	;.set device to be console
	STA	DEVNO
	CALL	CRLF
	LXI	H,ERRMSG	;.output message string
	CALL	TXTLOP
	POP	PSW	;.get error code number back
	MOV	L,A	;.put it into HL
	MVI	H,0
	CALL	INTOHL	;.print it in decimal
	MVI	A,'@'	;.print address mark
	CALL	OUTTO
	POP	H	;.get I2L PC at time of error
	CALL	HEXO	;.output address in hex
	CALL	CRLF
	JMP	ABORT	;.and abort the run

;.
;.	*** I2L Interpreter Loader module ***
;.
;.  Load format is normal ASCII interpreted as follows:
;.
;.	; - relative load address follows, 4 chars
;.		the PC is reset to this value
;.
;.	* - relative address follows, 4 chars to be
;.		relocated and stored at current PC
;.
;.	^ - relative address follows, 4 chars.  This
;.		address should be modified to contain
;.		the current PC value.
;.
;.  After any of the above, there may follow 2 chars of hex for
;.  each normal byte for any number of bytes.
;.
;.	$ - End of load:
;.
;.  Comments in the input are delineated by \
;.

LODST:	CALL	SETSTK	;set up the stack pointer
	SUB	A
	STA	DEVNO	;select the console
	CALL	INIDEV	;initialize console
;**	CALL	INODEV
	LXI	H,DFCB+1
	MOV	A,M
	CPI	' '
	JZ	LODST2	;no filename
	CALL	RTEST
	MVI	A,3
	STA	LODDEV
	LXI	H,RBFR	;make sure read buffer pointer inited
	SHLD	RPTR
	LXI	H,DFCB
	LXI	D,RFCB
	LXI	B,16
	CALL	MOVE
	LXI	H,RFCB+9
	MOV	A,M
	CPI	' '
	JNZ	LODST1	;has filetype
	MVI	M,'I'
	INX	H
	MVI	M,'2'
	INX	H
	MVI	M,'L'
LODST1:	LXI	H,DFCB+16+1
	CALL	RTEST
LODST2:	LHLD	BASADR	;.get base address to load starting at
	INX	H	;bump by one to preserve return from
			;  main (the opcode 0 for EXIT)
	SHLD	TEMP	;.save for relocation
	MVI	A,0
	STA	RERUNF	;.clear the rerun flag
	LDA	LODDEV	;see if filename already in read FCB
	ORA	A
	JZ	LODST3	;..no, so get filename
	CALL	FNDIT	;open the file if possible
	JNC	LODST4	;show "LOADING" message
LODST3:	CALL	CRLF	;.advance a line
	LXI	H,LODMS0 ;.output a prompt
	CALL	TXTLOP
	MVI	A,6	;get filename
	STA	LODDEV	;.save for later
	STA	DEVNO
	CALL	INIDEV	;.initialize load device
LODST4:	MVI	A,0	;.send loading message to console
	STA	DEVNO
	CALL	CRLF
	LXI	H,LODMS1
	CALL	TXTLOP
	CALL	CRLF
	LDA	LODDEV	;.set load device as current device
	STA	DEVNO

;.  Fall into main load loop

;.  Begin the load proper

LODLP:	CALL	BYTEIN	;get first char if not hex, first
			;byte if hex chars
	JC	LODLP2	;..not hex, so check for control char
	CALL	STOPC	;store byte at I2L PC, bump PC
	JMP	LODLP	;and do again until stopped

LODLP2:	CPI	';'	;.relative load addr?
	JZ	NEWADR
	CPI	'^'	;.relative addr that needs fixup?
	JZ	FIXUP
	CPI	'*'	;.relative addr?
	JZ	RELAD
	CPI	'$'	;.end of load?
	JZ	ENDUP
LFAIL:	MVI	A,LODERR ;.get error code
	JMP	ERRORP	;.and inform user

NEWADR:	CALL	READR	;.fetch new address
	LHLD	ADRES	;.into PC
	SHLD	PC
LODLP1:	CALL	BYTEIN	;.try to get a hex byte
	JC	LODLP2	;.jump if not hex
	CALL	STOPC	;.otherwise store it
	JMP	LODLP1	;.and go back for more

RELAD:	CALL	READR	;.get and relocate the address
	LDA	ADRES	;.and store it via the PC
	CALL	STOPC
	LDA	ADRES+1
	CALL	STOPC
	JMP	LODLP1	;.and back for more

FIXUP:	CALL	READR	;.get the address to fix up
	LHLD	ADRES	;.into HL
	LDA	PC	;.now store the current PC into that
			;.  address
	MOV	M,A
	INX	H
	LDA	PC+1
	MOV	M,A
	JMP	LODLP1	;.and back for more

ENDUP:	LHLD	PC	;.set STKLO to reflect space left
	SHLD	STKLO
	LXI	H,START	;.set up start addr to point to program
	SHLD	ENTER+1	;  loaded
	LDA	AUTO
	ORA	A
	JNZ	EXEC
	LXI	H,LENMS	;print program length
	CALL	CONPRT
	LHLD	STKLO
	PUSH	H
	MOV	L,H	;divide by 256 (give length in pages)
	MVI	H,0
	SHLD	REG1
	MVI	A,0	;print length on console (device 0)
	STA	DEVNO
	CALL	INTO	;print length of program in pages (decimal)
	LXI	H,PAGMS	;finish length message
	CALL	CONPRT
	POP	H
	INR	H
	MVI	L,0
	CALL	HEXO	;print program end address
	LXI	H,EPAMS
	CALL	CONPRT
	LXI	H,SAVMS	;print save file query
	CALL	CONPRT
	CALL	CONGET
EXEC:	LXI	H,START	;..not a Control C, so start the program
	PCHL

;.  Here to read an address from the I2L input file and
;.  relocate it.
;.

READR:	CALL	BYTEIN	;.read high order
	JC	LFAIL	;.error if not hex
	STA	ADRES+1	;.store it
	CALL	BYTEIN	;.read low order
	JC	LFAIL	;.error if not hex
	STA	ADRES	;.store it
	LHLD	ADRES	;.get address input
	XCHG		;.  into DE
	LHLD	TEMP	;.get relocation constant
	DAD	D	;.add it in to relocate address
	SHLD	ADRES	;.store relocated address back
	RET		;.and return

;.  Here to store the byte in the A register into memory
;.  via the I2L PC.
;.

STOPC:	LHLD	PC	;.get current PC value
	MOV	M,A	;.store byte
	INX	H	;.bump PC
	SHLD	PC	;.store updated PC back
			;and fall into heap overflow routine

;
;	TESTER - Make sure value in HL < value in HEAPHI
;
;  INPUT-
;	Value in HL
;	Value in HEAPHI (maximum allowed value of heap pointer)
;
;  OUTPUT-
;	HL = HEAPHI - HL
;	Return normally if value in HL <= value in HEAPHI
;	Output error 2 (out of room) otherwise
;

TESTER:	XCHG
	LHLD	HEAPHI	;get max allowed value of HP
	CALL	CMPU
	RNC		;..DE<=HL, so HL<=HEAPHI
	MVI	A,RMERR
	JMP	ERRORP	;put I2L PC back into reg BC, then
			;  report error since HL > HEAPHI

;.
;.  Here to input a byte from 2 hex digits from input device.
;.

BYTEIN:	CALL	INHEX	;.try to get a hex digit
	RC		;...not hex
	ADD	A	;.place it in high order nibble
	ADD	A
	ADD	A
	ADD	A
	PUSH	PSW	;.save it
	CALL	INHEX	;.get another hex digit
	JC	LFAIL	;...not hex
	MOV	D,A	;.save low order nibble
	POP	PSW	;.get high order back
	ORA	D	;.form a byte
	RET		;.and return with byte in reg A and
			;.  carry flag cleared.

INHEX:	CALL	ICN	;.get a char from input device
INHEX3:	CALL	UPC
	CPI	'0'	;.is it < '0'?
	JM	NOTHEX	;...no, so not hex
	CPI	'A'	;.is it < 'A'?
	JP	LETTER	;...no, may be a letter
	CPI	'9'+1	;...yes, but is it < '9'?
	JP	NOTHEX	;...no, so it is not hex
	ANI	0FH	;...yes, so form binary and return with
	RET		;.  carry flag cleared

LETTER:	CPI	'F'+1	;.is char <= 'F'?
	JP	NOTHEX	;...no, so not hex
	ANI	7	;...yes, so 'A'-'F', keep only 3 bits
	ADI	9	;.and form binary 10 - 15
	RET		;.and return with carry clear

NOTHEX:	STC		;.set carry flag to show not hex
	RET

;.
;.  Here to get a char during loading.
;.

ICN:	CALL	LODGET	;.get a char
	CPI	22H	;.ignore if funny (CR, LF, etc.)
	JC	ICN
	CPI	'\'	;.comment?
	RNZ		;...no, so return with char in reg A
COMMNT:	CALL	LODGET	;.comment ,so wait for '\'
	CPI	'"'	;.if text, skip text string
	JZ	TXTIGN
	CPI	'\'	;.hit matching '\' yet?
	JNZ	COMMNT	;...no, so keep ignoring chars
	JMP	ICN	;...yes, so get another char

TXTIGN:	CALL	LODGET	;.inside text string, so wait for the
	CPI	'"'	;.  ending '"'
	JNZ	TXTIGN
	JMP	COMMNT	;.string finished, but still inside
			;.  the comment

;.  Here to get a character from the load device.
;.

LODGET:	LDA	LODDEV	;.make load device
	STA	DEVNO	;.  the current device
	JMP	GETC

;
;	GETC - Get one ASCII character
;
;  INPUT-
;	DEVNO contains the device number to get char from
;  OUTPUT-
;	ASCII char in reg A and mem loc CHAR
;

GETC:	LDA	DEVNO	;get the device number
	ORA	A	;set flags
	JZ	GETC1	;..its the console, so get char with
			;  keyboard editing
	CALL	INPB	;..not console, so use normal input routine
	STA	CHAR	;save the character in mem loc CHAR
	RET

GETC1:	LDA	INBUPT	;get input buffer pointer
	CPI	0FFH	;anything in buffer?
	JZ	GETC2	;..no, so get another line
	PUSH	H	;preserve registers
	PUSH	D
	MOV	E,A
	MVI	D,0
	LDA	NUMCAR	;see if pointing to last char+1
	CMP	E
	MOV	A,E	;put pointer back in reg A
	JNZ	GETC1A	;..no, there are still chars in the buffer
	MVI	A,CR	;..yes, so force a carriage return
	JMP	GETC1B

GETC1A:	INR	A	;bump buffer pointer
	STA	INBUPT
	LXI	H,INBUF	;point to beginning of buffer
	DAD	D	;HL points to char in buffer
	MOV	A,M	;get the char
GETC1B:	STA	CHAR	;save it
	POP	D	;restore registers
	POP	H
	CPI	CR	;is the char a carriage return? (end of line)
	RNZ		;..no, so just return
	MVI	A,0FFH	;..yes, so flag buffer empty
	STA	INBUPT
	LDA	CHAR	;get character again
	RET

GETC2:	SUB	A	;clear the buffer
	STA	INBUPT
	LDA	PROMPT	;get the prompt char
	ORA	A
	JZ	GETC4	;no prompt if zero
	CALL	OUTTO	;and prompt the user
GETC4:	PUSH	H	;preserve all registers
	PUSH	D
	PUSH	B
	PUSH	PSW
	LXI	D,CIBUF	;use CP/M's buffered input function
	MVI	C,BUFINP
	CALL	CPMIO
	MVI	A,LF	;do a line feed after the CR
	CALL	TVOUT
	POP	PSW	;restore all registers
	POP	B
	POP	D
	POP	H
	JMP	GETC1	;and return with the first char from the buffer

;
;	INPB - Input a byte into reg A from device DEVNO
;

INPB:	MVI	A,0	;function code 0
	CALL	HAND	;go get it
	RET

;
;	OUTTO - Output a byte from reg A to device in DEVNO
;

OUTTO:	PUSH	PSW	;save character for later
	STA	ACPARM	;pass the byte as the parameter
	MVI	A,1	;function code 1
	CALL	HAND	;go send the byte
	POP	PSW	;get the char back again
	RET

;
;	CLODEV - Close output device DEVNO
;

CLODEV:	PUSH	PSW	;save reg A
	MVI	A,2	;function code 2
	CALL	HAND	;go do it
	POP	PSW	;restore reg A
	RET

;
;	INIDEV - Open input on device DEVNO
;

INIDEV:	PUSH	PSW	;save reg A
	LDA	DEVNO	;see if device is console
	ORA	A	;set flags
	JNZ	INID1	;..not console
	MVI	A,0FFH	;..console, so reset input buffer
	STA	INBUPT
INID1:	MVI	A,3	;function code 3
	CALL	HAND	;go do it
	POP	PSW	;restore reg A
	RET

;
;	INODEV - Open output on device DEVNO
;

INODEV:	PUSH	PSW	;save reg A
	MVI	A,4	;function code 4
	CALL	HAND	;go do it
	POP	PSW	;restore reg A
	RET

;
;	HAND - Routine to pass control to appropriate device
;

HAND:	PUSH	H	;preserve registers
	PUSH	D
	PUSH	B
	LXI	H,HAND1	;get addr of routine we want to return to
	PUSH	H	;put it on the stack as a return addr
	ADD	A	;form function number * 18
	MOV	B,A
	ADD	A
	ADD	A
	ADD	A
	ADD	B
	MOV	B,A	;and save it
	LDA	DEVNO	;get the device number
	CPI	DEVMAX	;compare with maximum allowed
	JNC	HAND2	;..illegal device number
	ADD	A	;calc (DEVNO)*2
	ADD	B	;add in function offset
	LXI	H,IOTAB	;calc table address for proper function
			;  and device
	MOV	C,A
	MVI	B,0
	DAD	B	;HL=addr containing addr of desired routine
	MOV	E,M	;get addr of desired routine in DE
	INX	H
	MOV	D,M
	XCHG		;HL=addr of desired routine
	LDA	ACPARM	;get parameter into reg A
	PCHL		;do the function

HAND1:	STA	ACPARM	;save the parameter
	POP	B	;restore registers
	POP	D
	POP	H
	RNC		;return if no error
;
	LDA	DEVNO	;get device number
	STA	BADDEV	;save it as bad device
	MVI	A,IOERR	;indicate IO error
	JMP	ERROR

HAND2:	POP	H	;remove pointer to HAND1
	POP	B	;restore registers
	POP	D
	POP	H
	MVI	A,DEVERR ;indicate bad device number
	JMP	ERROR

;
;	NULDEV - dummy driver for unused devices
;

NULDEV:	SUB	A	;clear carry flag (always OK)
	MVI	A,EOFCHR ;return EOF char if input
	RET

;
;	GETNO - get a signed number and put it into REG1
;
;  INPUT-
;	Decimal number from input device
;	  No sign means it's positive
;	  Minus sign means it's negative
;	  Chars allowed are '-' '0' though '9'
;	  and '_'

GETNO:	SUB	A
	STA	SFLG	;clear the sign flag
	STA	INPFLG	;indicate no input yet
	LXI	H,0
	SHLD	REG1	;clear temporary registers
	CALL	GETC	;get a char
	CPI	'-'	;is it a minus sign?
	JNZ	GETN2	;..no
	MVI	A,80H	;..yes, so set sign flag
	STA	SFLG
GETN1:	CALL	GETC	;get another char
GETN2:	CPI	'_'	;ignore any underlines
	JZ	GETN1
	CPI	EOFCHR
	JZ	GETN4
	LDA	CHAR	;get the digit again
	SBI	'0'	;change from ASCII to value
	JC	GETN3	;..not a digit, so this is the end of
			;  the number
	CPI	0AH
	JNC	GETN3
	LHLD	REG1	;get interim results so far
	MOV	D,H
	MOV	E,L	;copy into DE
	DAD	H	;*2
	DAD	H	;*4
	DAD	D	;*5
	DAD	H	;*10
	MOV	E,A	;add value of this digit
	MVI	D,0
	DAD	D
	SHLD	REG1	;save interim result
	MVI	A,0FH	;indicate we have some now
	STA	INPFLG
	JMP	GETN1	;and check it out

GETN3:	LDA	INPFLG	;did we get a number at all?
	ORA	A	;set flags
	JZ	GETNO	;..no, so try again
GETN4:	LDA	SFLG	;was the number negative?
	ORA	A
	LHLD	REG1
	CNZ	NEGG	;..yes, so negate it
	SHLD	REG1
	RET

;
;  INTO - print decimal number from REG1 (REG1 is destroyed)
;  INTOHL - print decimal number from HL (HL is destroyed)
;
;  INPUT-
;	REG1 contains 16 bit number to be output
;	DEVNO contains device number of device number is output to
;
;  Both routines destroy Flags, A, D, E, H and L
;

INTO:	LHLD	REG1	;get the number into HL
INTOHL:	SUB	A	;zero reg A
	STA	SFLG	;set suppress flag to suppress leading
			;  zeroes
	ORA	H	;test sign bit
	JP	MORT	;..not negative, so output as is
	MVI	A,'-'	;output a minus sign
	CALL	OUTTO
	CALL	ABSA	;convert number in HL to positive
MORT:	LXI	D,0
	PUSH	D	;put 0 on stack (done flag)
	LXI	D,10
	PUSH	D
	LXI	D,100
	PUSH	D
	LXI	D,1000
	PUSH	D
	LXI	D,10000	;get binary value of 10000 into DE
	PUSH	D	;and put it on the stack
MORT0:	POP	D	;get ten to some power
	MOV	A,D	;see if DE=0 (routine done)
	ORA	E
	JZ	INTOX	;..routine done
	PUSH	D	;put ten to power back on stack
	SUB	A	;negate the value in DE
	SUB	E
	MOV	E,A
	MVI	A,0	;zero reg A without disturbing carry
	SBB	D
	MOV	D,A
			;HL=number to be output
			;DE=-(ten to some power)
	MVI	A,'1'	;preset counter to ASCII '1'
MORT1:	DAD	D
	JNC	ZCHK	;..result is zero or negative
	INR	A	;update counter
	JMP	MORT1	;and do it again

ZCHK:	MOV	E,A	;save the counter
	MOV	A,H
	ORA	L	;see if result is zero
	MOV	A,E	;get the counter back
	JZ	ZCHK1	;..yes, so leave counter alone
	DCR	A	;..no, so reduce count by one
	POP	D	;get ten to same power again
	PUSH	D	;put it back on the stack
	DAD	D	;make result non-negative
ZCHK1:	POP	D	;clear present ten to power from stack
	SHLD	REG1	;save result for later
	MOV	L,A	;save the digit
	LDA	SFLG	;if the suppress flag#0 or digit#0, set
	ORA	L	;  flag for no suppression and print
			;  digit
	SUI	'0'	;remove ASCII bias from char to get value
	STA	SFLG
	MOV	A,L	;get the ASCII digit again
	CNZ	OUTTO	;..output the digit
	LHLD	REG1	;get the result again
	JMP	MORT0	;keep it up till all digits done

INTOX:	MVI	A,'0'	;convert the last digit to ASCII
	ADD	L
	CALL	OUTTO	;and output it
	RET

;
;  HEXO-
;
;  Routine to output 16 bit value in HL as 4 hex digits
;

HEXO:	MOV	A,H	;get high byte
	CALL	HEX2	;output the 2 digits
	MOV	A,L	;get low byte
	CALL	HEX2	;output the two digits
	RET
;
;  Convert the byte to two hex chars and output them
;
HEX2:	PUSH	PSW	;save the value for now
	RRC		;move high nibble into low
	RRC
	RRC
	RRC
	CALL	HEX3	;convert the four bits
	POP	PSW	;get the byte back again
;
;  Convert low nibble into Hex character and output it
;
HEX3:	ANI	0FH	;keep only the low nibble
;
;  convert 4 bit value to ASCII
;
	ADI	90H
	DAA
	ACI	40H
	DAA
	CALL	OUTTO	;print the ASCII char
	RET
;=======================================================
;
;	CP/M IO routines
;
FCBADR:	DW	0	;holds the FCB address of the file
			;  to be opened
;
;*******************************************************
;
;  Get file name and find file on the disk
;
;*******************************************************
;
GETCPM:	LXI	H,CPMQRY ; query for CPM file name
	CALL	CONPRT	;print query on console
	CALL	CONGET	;get the response (string with editing)
	MVI	A,0FFH	;reset input buffer pointer (to show
	STA	INBUPT	;  buffer is empty next time it is used)
	LXI	H,INBUF	;point to beginning of input buffer
	LDA	NUMCAR	;get number of valid chars in buffer
	ADD	L	;calc addr of last valid char
	MOV	L,A
	MOV	A,H
	ACI	0
	MOV	H,A
	INX	H	;point to cell following last valid char
	MVI	M,0	;put a null there for SETFCB
	LHLD	FCBADR	;get FCB address
	XCHG
	LXI	H,INBUF	;point to beginning of file name
	CALL	MAKFCB	; create the FCB
	LHLD	FCBADR	; search for the file
	XCHG
	MVI	C,SERFIL
	CALL	CPMIO
	CPI	0FFH	; check for file not found
	RET		;zero flag set means no such file
;
;*******************************************************
;
;  Open CP/M file for output
;
;*******************************************************

OPNOUT:	LXI	H,WFCB	;set up for write
	SHLD	FCBADR
	LXI	H,OUTMS	;prompt for output file name
	CALL	CONPRT
	CALL	GETCPM	;get the file name and try to find it
	JZ	FILEOK	;..not found
	LXI	H,FILTHER ; inform user that the file exists
	CALL	CONPRT	;  and ask if OK to delete old file
LOOP2:	CALL	CONGET	; get response
	ANI	5FH	;strip parity, ignore upper/lower case
	CPI	'Y'	; check for Y
	MVI	A,0FFH	;clear the buffer pointer
	STA	INBUPT
	JNZ	OPNOUT	; if not then ask for file again
OPNOUT1:		;delete existing file (if any) then open file for
	LXI	D,WFCB	;  output
	MVI	C,DELFIL ; delete existing file
	CALL	CPMIO

FILEOK:	LXI	D,WFCB	; create the file
	MVI	C,MAKFIL
	CALL	CPMIO
	STA	OOER3	;save open output floppy error
	STA	OOER7	;also unprompted floppy
	CPI	0FFH	; check for no room
	JNZ	XFER	;..room OK
	LXI	H,NOROOM ; give no room message
	CALL	CONPRT
	JMP	OPNOUT	;allow user to try again on other disk

;  Initialize output buffer

XFER:	LXI	D,DDMA	; set dma addr
	MVI	C,SETDMA
	CALL	CPMIO

	LXI	H,WBFR	;reset write buffer pointer
	SHLD	WPTR
	SUB	A	;clear carry (IO error) flag
	RET

;
;*******************************************************
;
;  Open a CP/M file for input
;
;*******************************************************

OPNIN:	LXI	H,RFCB	;set up for read
	SHLD	FCBADR
	LXI	H,INMS	;prompt for input file name
	CALL	CONPRT
	CALL	GETCPM	;get the file name and try to find it
	JNZ	FNDIT	;..found the file
	LXI	H,NOTHER ;..file not there, so inform user
	CALL	CONPRT
	JMP	OPNIN	;and try again

FNDIT:	MVI	C,OPNFIL ;open read file
	LXI	D,RFCB
	CALL	CPMIO
	STA	OIER3	;save error byte for open floppy input
	STA	OIER7	;also for unprompted floppy
	CPI	0FFH	;check for file not there
	JZ	OPNER	;..tell user can't find file
	CALL	REDBFR	;get the first sector's worth into the
			;  buffer
	SUB	A	;clear carry (IO error) flag
	RET

OPNER:	LXI	H,NOTHER ;report file not there
	CALL	CONPRT
	STC		;set carry flag (IO error)
	RET

;
;*******************************************************
;
;  Write a byte to CP/M
;
;*******************************************************

BYTOCP:	LHLD	WPTR	;get write buffer pointer
	MOV	M,A	;put the char into the buffer
	INX	H	;update the buffer pointer
	SHLD	WPTR
	MOV	A,L	;see if buffer is full
;	CPI	LOW (WBFR+80H)
	CPI	(WBFR+80H) AND 255
	JZ	WRIBFR	;..yes, so put buffer on disk
	SUB	A	;clear carry (IO error) flag
	RET
;
;  Copy the write buffer into the DDMA buffer then write
;  the DDMA buffer onto the disk
;
WRIBFR:	PUSH	B	;preserve BC
	LXI	B,DDMA	;point to beginning of DDMA buffer
	LXI	H,WBFR	;  and beginning of write buffer
	SHLD	WPTR	;reset write buffer pointer
WRI1:	MOV	A,M
	STAX	B
	INX	H
	INX	B
	SUB	A
	ADD	C
	JNZ	WRI1	;..more bytes to copy
	POP	B	;restore BC
	CALL	WRITCP	;write the DDMA buffer onto disk
	RET
;
;*******************************************************
;
;  Read a byte from CP/M
;
;*******************************************************

BYFMCP:	LHLD	RPTR	;get the read buffer pointer
	MOV	A,M	;get the char
	INX	H	;update the buffer pointer
	SHLD	RPTR
	MOV	H,A	;save the char for now
	MOV	A,L	;check to see if finished buffer
;	CPI	LOW (RBFR+80H)
	CPI	(RBFR+80H) AND 255
	JZ	REDBFR	;..yes, so get another bufferful
	SUB	A	;clear IO error (carry) flag
	MOV	A,H	;get the char again
	RET
;
;  Read the next bufferful into the DDMA buffer, then
;  copy from DDMA buffer into read buffer
;
REDBFR:	PUSH	H	;save the character
	PUSH	B	;save BC
	CALL	READIT	;get a bufferful into DDMA buffer
	LXI	H,RBFR	;point to beginning of read buffer
	SHLD	RPTR	;reset read buffer pointer
	LXI	B,DDMA	;point to beginning of DDMA buffer
RED1:	LDAX	B
	MOV	M,A
	INX	H
	INX	B
	SUB	A
	ADD	C
	JNZ	RED1	;..more bytes to copy
	POP	B	;restore BC
	POP	H	;get the char again (in H)
	MOV	A,H	;put char back into reg A
	RET

;
;*******************************************************
;
;  Close CP/M file
;
;*******************************************************

CLSCPM:	LHLD	WPTR
	MVI	M,1AH	; append ^Z to end of file
	CALL	WRIBFR	; write last sector
	LXI	D,WFCB
	MVI	C,CLSFIL ; close the file
	CALL	CPMIO
	STA	CER3	;save error code in close floppy
	STA	CER7	;also in unprompted floppy close
	SUB	A	;clear carry (IO error) flag
	RET
;
;*******************************************************
;
;  Write a sector of the CP/M file
;
;*******************************************************

WRITCP:	LXI	D,WFCB	; write a sector
	MVI	C,WRITSEC
	CALL	CPMIO
	ORA	A	; check for errors
	RZ		; return if no errors
	STA	OER3	;save error code for floppy output
	STA	OER7	;also in unprompted floppy output
	LXI	H,ERRMS1
	CPI	1
	JZ	ERRWRT
	LXI	H,ERRMS2
	CPI	2	; check for end of disk data
	JZ	ERRWRT
	LXI	H,ERRMS3
ERRWRT:	CALL	CONPRT	; print error message
	LXI	D,WFCB	; close what there is of the file
	MVI	C,CLSFIL
	CALL	CPMIO
	STC		;set carry (IO error) flag
	RET

;
;*******************************************************
;
;  Read a sector from CP/M
;
;*******************************************************

READIT:	LXI	D,RFCB	;read a sector from read file
	MVI	C,READSEC
	CALL	CPMIO
	ORA	A	;check end code
	JNZ	STOP	;..file emptied
	RET

STOP:	STA	IER3	;store error code for floppy input
	STA	IER7	;also in unprompted disk input
	LXI	H,MTMSG	;tell user file is empty
	CALL	CONPRT
	RET		;and return normally

;
;  Get a byte from the console (without destroying DEVNO)
;
CONGET:	LDA	DEVNO	;save old device number
	PUSH	PSW
	SUB	A	;make console I/O device
	STA	DEVNO
	CALL	GETC
	MOV	L,A	;save char in reg L
	POP	PSW	;restore old device number
	STA	DEVNO
	MOV	A,L	;put char back into reg A
	RET

;
;  Print a string on the console (wihout destroying DEVNO)
;
CONPRT:	LDA	DEVNO	;save old device number
	PUSH	PSW
	SUB	A	;set console to be I/O device
	STA	DEVNO
	CALL	TXTLOP	;print the string
	POP	PSW	;restore old device number
	STA	DEVNO
	RET

;
;	CP/M IO Messages
;
LENMS:	DB	CR,LF,'Program length =',' '+80H
PAGMS:	DB	' pages [0100','-'+80H
EPAMS:	DB	']',CR,LF+80H
SAVMS:	DB	CR,LF,'Press Control C to save run-time version on disk',CR,LF
	DB	'or RETURN to execute',CR,LF+80H
INMS:	DB	CR,LF,'Inpu','t'+80H
OUTMS:	DB	CR,LF,'Outpu','t'+80H
CPMQRY:	DB	' filename:',' '+80H
NOROOM:	DB	CR,LF,'******  No room on disk for a new file  ******',CR,LF+80H
ERRMS1:	DB	CR,LF,'Error in extending fil','e'+80H
ERRMS2: DB 	CR,LF,'End of disk dat','a'+80H
ERRMS3: DB	CR,LF,'No more directory spac','e'+80H
FILTHER: DB	CR,LF,'File already exists, OK to replace? (Y = yes)',' '+80H
MTMSG:	DB	CR,LF,'End of File',CR,LF+80H
NOTHER:	DB	CR,LF,'File not on disk',CR,LF+80H

;--------------------------------------------------------------
;
;	Routine to make an FCB from input of the form:
;		A:FILENAME.EXT
;	if a drive specification is made, the drive number
;	is placed in the first byte of the FCB buffer
;
;	input:	HL = address of input string terminated by null
;		or bit 7 of last character set. Leading spaces
;		are skipped.
;		DE = address of FCB to setup
;
MAKFCB:	PUSH	B
	XCHG
	PUSH    H
	MVI	B,12
MKFC1:	MVI	M,' '
	INX	H
	DCR	B
	JNZ     MKFC1
	MVI	B,21
	XRA	A
MKFC2:	MOV	M,A
	INX	H
	DCR	B
	JNZ     MKFC2
	POP	H
	MOV	M,A
	DCR	A
	STA	EFLG
	XCHG
MKFC3:	CALL	MGET
	CPI	' '
	INX	H
	JZ	MKFC3
	CALL	MGET
	CPI	':'
	DCX	H
	JNZ	MKFC4
	CALL	MGET
	SBI	'@'
	STAX	D
	INX	H
	INX	H
MKFC4:	INX	D
	MVI	B,8
	CALL	MKFC6
	CALL	MGET
	CPI	'.'
	JNZ	MKFC5
	INX	H
	MVI	B,3
	CALL	MKFC6
MKFC5:	POP	B
	RET

MKFC6:	CALL	MGET
	CPI	'*'
	MVI	C,'?'
	JZ	MKFC8
	CPI	'.'
	MVI	C,' '
	JZ	MKFC9
	CPI	' '+1
	RC
	CPI	','
	RZ
	CPI	'_'+1
	RNC
	CPI	'.'
	RZ
	CPI	':'
	JC	MKFC7
	CPI	'>'+1
	RC
MKFC7:	STAX	D
	INX	H
	INX	D
	DCR	B
	JNZ     MKFC6
	RET

MKFC8:	INX	H
MKFC9:	MOV	A,C
MKFC10:	STAX	D
	INX	D
	DCR	B
	JNZ     MKFC10
	RET

MGET:	MOV	A,M
	PUSH	H
	LXI	H,EFLG
	ANA	M
	JZ	MGT2
	JP	MGT1
	ANI	7FH
	INR	M
MGT1:	CALL	UPC
MGT2:	POP	H
	RET

EFLG:	DB	0FFH

;
;	Make A uppercase
;
UPC:	CPI	'a'
	RC
	CPI	'z'+1
	RNC
	XRI	20H
	RET

;	Call CBIOS function
;
;	Inputs: BC, DE  Fn#: A
;
;	Works for any CP/M 2 BIOS function and
;	CP/M 3 console/list BIOS functions only

CBIOS:	PUSH	D
	MOV	E,A
	MVI	D,0
	LHLD	0001H
	DCX	D
	DAD	D
	DAD	D
	DAD	D
	POP	D
	PCHL

;	END of CP/M IO routines

;PSEUDO DEVICE HANDLER ROUTINES
; THE PSEUDO DEVICE IS A CIRCULAR MEMORY BUFFER WHICH CAN BE
; READ OR WRITEN TO. SIMPLIFIES STRING PARSING AND NUMBER
; CONVERSION

PSUSIZ	EQU	256		;SIZE OF BUFFER
PSUBUF:	DS	PSUSIZ		;PSEUDO DEVICE BUFFER
PSUEND	EQU	$
PSIRED:	DW	PSUBUF		;READ POINTER
PSIWRT:	DW	PSUBUF		;WRITE POINTER
PSISTR:	DW	PSUBUF		;START POINTER

;OPEN PSEUDO DEVICE FOR OUTPUT

PSOPNO:	LXI	H,PSUBUF
	SHLD	PSIWRT		;ZERO WRITE POINTER
	SHLD	PSISTR		;ZERO START POINTER

;OPEN PSEUDO DEVICE FOR INPUT

PSOPNI:	LHLD	PSISTR		;SET READ POINTER TO START
	SHLD	PSIRED
	ORA	A
	RET

;OUTPUT A BYTE TO PSEUDO DEVICE

PSOUT:	LHLD	PSIWRT		;GET WRITE POINTER
	MOV	M,A		;STORE CHAR IN BUFFER
	PUSH	PSW
	CALL	INCWRT		;INCREMENT WRITE POINTER
	POP	PSW
	ORA	A
	RET

;INPUT A BYTE FROM PSEUDO DEVICE

PSIN:	LHLD	PSIWRT
	XCHG
	LHLD	PSIRED		;GET READ POINTER
	CALL	CMPU		;READ=WRITE IF EMPTY
	JNZ	PSIN1		;SKIP IF MORE TO READ
	MVI	A,EOFCHR	;SIGNAL END OF FILE
	ORA	A		;FLAG NO ERROR
	RET

PSIN1:	PUSH	H
;	LHLD	PSIRED
	INX	H		;ADVANCE POINTER
	SHLD	PSIRED
	LXI	D,PSUEND	;WRAP AROUND?
	CALL	CMPU
	JNZ	PSIN2		;SKIP IF NOT
	LXI	H,PSUBUF	;START OVER
	SHLD	PSIRED
PSIN2:	POP	H
	MOV	A,M		;READ BYTE
	ORA	A		;FLAG NO ERROR
	RET

;ROUTINE TO INCREMENT WRITE POINTER AND HANDLE WRAP AROUND
;AND COLLISIONS

INCWRT:	LHLD	PSIWRT		;INCREMENT WRITE POINTER
	INX	H
	SHLD	PSIWRT
	LXI	D,PSUEND	;WRAP AROUND?
	CALL	CMPU
	JC	INCWT1		;SKIP IF NOT
	LXI	H,PSUBUF	;THEN START OVER
	SHLD	PSIWRT
INCWT1:	LHLD	PSISTR		;CAUGHT START POINTER?
	XCHG
	LHLD	PSIWRT
	CALL	CMPU
	JNZ	INCWT2		;SKIP IF NOT
	LHLD	PSISTR		;INCREMENT START POINTER
	INX	H
	SHLD	PSISTR
	LXI	D,PSUEND	;WRAP AROUND?
	CALL	CMPU
	JC	INCWT2		;SKIP IF NOT
	LXI	H,PSUBUF	;START OVER
	SHLD	PSISTR
INCWT2:	LHLD	PSIRED		;CAUGHT READ POINTER?
	XCHG
	LHLD	PSIWRT
	CALL	CMPU
	JNZ	INCWT3		;SKIP IF NOT
	LHLD	PSIRED		;INCREMENT READ POINTER
	INX	H
	SHLD	PSIRED
	LXI	D,PSUEND	;WRAP AROUND?
	CALL	CMPU
	JC	INCWT3		;SKIP IF NOT
	LXI	H,PSUBUF	;START OVER
	SHLD	PSIRED
INCWT3:	RET

;
;	MOVE - Move Block
;
;	HL = src, DE = dest, BC = #bytes
;

MOVE:	MOV	A,C
	ORA	B
	RZ
	MOV	A,M
	STAX	D
	INX	H
	INX	D
	DCX	B
	JMP	MOVE

;	HL=ADDR

RTEST:	MOV	A,M
	CPI	'-'
	RNZ
	INX	H
	MOV	A,M
	CALL	UPC
	CPI	'R'
	RNZ
	MVI	A,1
	STA	AUTO
	RET

;---------------------------------------------------------------
;
;  Turbo Pascal compatible terminal functions
;
;  DELAY  CURSOR  CLEAR  NORMAL  BOLD  CLREOL  INSLINE  DELLINE
;  CRTINIT  CRTEXIT
;

;$70 #48
;	DELAY(ms)
;
DELAY:	POP	D
	CALL	MS
	JMP	OPGO

;	DE = msec

MS:	MOV	A,E
	ORA	D
	RZ
	LHLD	SPEED
	DAD	H
	DAD	H
	DAD	H
MS02:	XTHL			; 19T
	XTHL			; 19T
	XTHL			; 19T
	XTHL			; 19T
	PUSH	H		; 11T
	POP	H		; 10T
	DCX	H		; 6T
	MVI	A,0		; 7T
	MOV	A,L		; 4T
	ORA	H		; 4T
	JNZ	MS02		; 10T
	DCX	D
	JMP	MS

CPBUF:	DS	16		; temp buffer

;$57 #23
;	CURSOR(col,row)		position cursor
;
CURSOR:	POP	D
	POP	H
	MOV	D,L
	LXI	H,TGXY
	MOV	A,M
	ORA	A
	JZ	OPGO		; empty string
	PUSH	D
	LXI	D,CPBUF
	LXI	B,16
	CALL	MOVE
	POP	D
	LDA	TXPOS
	MOV	C,A
	LDA	TXOFS
	ADD	D
	PUSH	D
	CALL	MCURS3
	POP	D
	LDA	TYPOS
	MOV	C,A
	LDA	TYOFS
	ADD	E
	CALL	MCURS3
	LXI	D,TGXYD
	LXI	H,CPBUF
	CALL	TSOUT
	JMP	CMLRET

;	send terminal string and delay

TSOUT:	PUSH	D
	MOV	B,M	; count
	PUSH	B
TSOUT1:	MOV	A,B
	ORA	A
	JZ	TSOUT2
	INX	H
	DCR	B
	MOV	A,M
	PUSH	H
	PUSH	B
	CALL	CRT1	; output char
	POP	B
	POP	H
	JMP	TSOUT1

TSOUT2:	POP	PSW	; count
	POP	H
	ORA	A
	RZ		; delay only if string not empty
	MOV	E,M
	MVI	D,0
	JMP	MS

MCURS3:	LXI	H,CPBUF
	MVI	B,0
	DAD	B
	XCHG
	LXI	H,TISBIN
	INR	M
	DCR	M
	JZ	MCURS4
	STAX	D
	RET

MCURS4:	DCX	D
	DCX	D
	LXI	H,MCURS8-1
	MVI	B,3
MCURS5:	INX	H
	MVI	C,'0'-1
MCURS6:	INR	C
	SUB	M
	JNC	MCURS6
	ADD	M
	PUSH	PSW
	MOV	A,C
	CPI	'0'
	JZ	MCURS7
	STAX	D
MCURS7:	INX	D
	POP	PSW
	DCR	B
	JNZ	MCURS5
	RET

MCURS8:	DB	100,10,1

;

CLS:	LXI	D,TCEOLD
	LXI	H,THOME
	CALL	TSOUT
	LXI	D,TCLSD
	LXI	H,TCLS
	CALL	TSOUT
	ORA	A		; clear carry
	RET

;$68 #40
;	CLEAR
;
VCLEAR:	CALL	CLS
	JMP	CMLRET

;$69 #41
;	CLREOL
;
CLREOL:	LXI	H,TCEOL
	JMP	VNORM1

;$71 #49
;	NORMAL
;
VNORM:	LXI	H,TLOVID
VNORM1:	LXI	D,TCEOLD
	CALL	TSOUT
	JMP	CMLRET

;$72 #50
;	BOLD
;
VBOLD:	LXI	H,THIVID
	JMP	VNORM1

;$73 #51
;	INSLINE
;
INSLIN:	LXI	H,TINSLN
INSLN1:	LXI	D,TCLSD
	CALL	TSOUT
	JMP	CMLRET

;$74 #52
;	DELLINE
;
DELLIN:	LXI	H,TDELLN
	JMP	INSLN1

;$75 #53
;	CRTINIT
;
CRTINI:	LXI	H,TINIT
	JMP	INSLN1

;$76 #54
;	CRTEXIT
;
CRTEXI:	LXI	H,TEXIT
	JMP	INSLN1

;
;=======================================================
;
;.	Message strings

ERRMSG:	DB	'I2L Error '
	DB	'#'+80H

LODMS0:	DB	'8080/Z-80 I2L  VER 1.2  2016-09-27'
	DB	CR,LF,'Copyright 1978,1979,1980 by Wayne Wall and Dana Trout'
	DB	CR,LF,'Usage: I2L filename[.I2L] [-R]',CR,LF
	DB	' '+80H

LODMS1:	DB	'*** Loading **'
	DB	'*'+80H

LOSPAC:	DB	0	;opcode for I2L EXIT (This is a dummy
			;  instruction used to terminate the
			;  user program.)

	END	100H

